class AppTables(object):
    def __repr__(self):
        return "<anvil.tables.{} object>".format(type(self).__name__)


class AbstractTableClass(object):
    _instead = None

    def __init__(self, *args, **kwargs):
        raise TypeError("Can't instantiate a {} object. Use {} instead.".format(type(self).__name__, self._instead))

    def __repr__(self):
        return "<anvil.tables.{} object>".format(type(self).__name__)

    def __dir__(self):
        # TODO should we keep this?
        # remove private attributes and methods from the dir
        return [key for key in object.__dir__(self) if (not key.startswith("_")) or key.startswith("__")]


class Table(AbstractTableClass):
    _instead = "app_tables.my_table"


class SearchIterator(AbstractTableClass):
    _instead = "app_tables.my_table.search()"


class Row(AbstractTableClass):
    _instead = "app_tables.my_table.add_row()"



def _hash_wrapper(*params):
    # this makes query objects cachable as keys of dictionaries
    def _mk_tuple(self):
        return tuple(getattr(self, param) for param in params)

    def __hash__(self):
        return hash(_mk_tuple(self))

    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        return _mk_tuple(self) == _mk_tuple(other)

    return __hash__, __eq__



import anvil.server

#!defModule(anvil.email)!1: "The `anvil.email` module contains functions for sending and receiving email in your Anvil app."

#!suggestAttr(anvil.email,send)!0:

#!defClass(anvil.email,SendFailure)!:
class SendFailure(anvil.server.AnvilWrappedError):
    pass

anvil.server._register_exception_type("anvil.email.SendFailure", SendFailure)

class DeliveryFailure(Exception):
    #!defMethod(_,message=None,smtp_code=554)!2: 
    # {anvil$helpLink: "/docs/email/sending_and_receiving#rejecting-email", $doc: "While handling an error, you can raise a DeliveryFailure exception to reject email delivery. Optionally, you may specify a message and SMTP error code with the rejection."} ["__init__"]
    def __init__(self, message=None, smtp_code=None):
        if message is None:
            super(DeliveryFailure, self).__init__()
        elif smtp_code is not None:
            message = "{}: {}".format(smtp_code, message)
        super(DeliveryFailure, self).__init__(message)
    #!defClass(anvil.email,DeliveryFailure)!:



#!defFunction(anvil.email,anvil.email.SendReport instance,[to=],[cc=],[bcc=],[from_address="no-reply"],[from_name=],[subject=],[text=],[html=],[attachments=],[inline_attachments=])!2:
# {
#   $doc: "Send an email",
#   anvil$helpLink: "/docs/email",
#   anvil$args: {
#     to: "The email recipient[s] in the 'To' field. Can be a string or list of strings.\n\nEach string can be a bare address (eg 'joe@example.com') or include a display name (eg 'Joe Bloggs <joe@example.com>').",
#     cc: "The email recipient[s] in the 'Cc' field. Can be a string or list of strings.\n\nEach string can be a bare address (eg 'joe@example.com') or include a display name (eg 'Joe Bloggs <joe@example.com>').",
#     bcc: "The email recipient[s] in the 'Bcc' field. Can be a string or list of strings.\n\nEach string can be a bare address (eg 'joe@example.com') or include a display name (eg 'Joe Bloggs <joe@example.com>').",
#     from_address: "The From: address from this email. Can be a bare address (eg 'joe@example.com') or include a display name (eg 'Joe Bloggs <joe@example.com>').\n\nIf no domain is specified, or the specified domain is not a legal sending domain for this app, the address will be replaced with a valid domain. So if you specify 'noreply', the email will come from 'noreply@your-app-domain.anvil.app'.",
#     from_name: "The name associated with the From: address for this email. (Only valid if the from_address is a bare email address.)",
#     subject: "The subject line for this email.",
#     text: "The plain-text (no HTML) content for this email. You must specify at least one of 'text' and 'html'.",
#     html: "The HTML content for this email. You must specify at least one of 'text' and 'html'.",
#     attachments: "A list of Media objects to send as attachments with this email.",
#     inline_attachments: "Inline that can be used in this email's HTML, for example in <img> tags. Must be a dictionary whose keys are IDs and values are Media objects. IDs can then be used in a message's HTML with 'cid:xxx' URIs.",
#   }
# } ["send"]
def send(**kw):
    return anvil.server.call("anvil.private.email.send.v2", **kw)

# NB no defFunction() here; this one is defined in the autocompleter
def handle_message(fn=None, require_dkim=False):
    def wrapper(fn):
        def handler(msg_dict):
            msg = Message(msg_dict)
            if require_dkim and not msg.dkim.valid_from_sender:
                raise DeliveryFailure("No valid DKIM signature for %s" % msg.envelope.from_address)
            fn(msg)
        return anvil.server.callable("email:handle_message")(handler)
    return wrapper(fn) if fn is not None else wrapper


@anvil.server.portable_class
class Address(object):

    #!defAttr()!1: {name:"address",type:"string",description:"The email address this object represents."}
    #!defAttr()!1: {name:"name",type:"string",description:"The name associated with the address this object represents."}
    #!defAttr()!1: {name:"raw_value",type:"string",description:"The full string value of this address."}
    def __init__(self, address):
        self.address = address['address']
        self.name = address['name']
        self.raw_value = address['raw']

    #!defClass(anvil.email,#Address)!:


@anvil.server.portable_class
class Message(object):
    #!defAttr()!1: {name:"from_address",type:"string",description:"The email address from which this message was sent, according to the SMTP envelope."}
    #!defAttr()!1: {name:"recipient",type:"string",description:"The email address that received this message.\n\nNote that this email address may not appear in any of the headers (eg if the email has been BCCed or blind forwarded)."}
    @anvil.server.portable_class
    class Envelope(object):
        def __init__(self, envelope):
            self.from_address = envelope['from']
            self.recipient = envelope['recipient']
    #!defClass(anvil.email.Message,#Envelope)!:

    #!defAttr()!1: {name:"valid_from_sender",type:"boolean",description:"Was this message signed by the domain in its envelope \"from\" address?"}
    #!defAttr()!1: {name:"domains",type:"list(string)",description:"A list of the DKIM domains that signed this message."}
    @anvil.server.portable_class
    class DKIM(object):
        def __init__(self, dkim):
            self.valid_from_sender = dkim['valid_from_sender']
            self.domains = dkim['domains']
    #!defClass(anvil.email.Message,#DKIM)!:


    #!defAttr()!1: {name:"to_addresses",pyType:"list(anvil.email.Address instance)",description:"The addresses this message was sent to."}
    #!defAttr()!1: {name:"from_address",pyType:"anvil.email.Address instance",description:"The address this message was sent from."}
    #!defAttr()!1: {name:"cc_addresses",pyType:"list(anvil.email.Address instance)",description:"The addresses this message was copied to."}
    @anvil.server.portable_class
    class Addressees(object):
        def __init__(self, addressees):
            self.to_addresses = [Address(a) for a in addressees.get('to',[])]
            self.from_address = Address(addressees['from'][0]) if 'from' in addressees else None
            self.cc_addresses = [Address(a) for a in addressees.get('cc',[])]
    #!defClass(anvil.email.Message,#Addressees)!:


    #!defAttr()!1: {name:"envelope",pyType:"anvil.email.Message.Envelope instance",description:"The sender and receipient of this email, according to the SMTP envelope."}
    #!defAttr()!1: {name:"dkim",pyType:"anvil.email.Message.DKIM instance",description:"Object describing whether this message was signed by the sending domain"}
    #!defAttr()!1: {name:"addressees",pyType:"anvil.email.Message.Addressees instance",description:"The addresses this email was sent from and to, according to the headers."}
    #!defAttr()!1: {name:"headers",type:"list",description:"All the headers in this email, as a list of (name,value) pairs."}
    #!defAttr()!1: {name:"text",type:"string",description:"The plain-text content of this email, or None if there is no plain-text part."}
    #!defAttr()!1: {name:"subject",type:"string",description:"The subject of this email, or None if there is no subject."}
    #!defAttr()!1: {name:"html",type:"string",description:"The HTML content of this email, or None if there is no HTML part."}
    #!defAttr()!1: {name:"attachments",pyType:"list(anvil.Media instance)",description:"A list of this email's attachments."}
    #!defAttr()!1: {name:"inline_attachments",pyType:"dict(string,anvil.Media instance)",description:"A dictionary of this email's inline attachments. Keys are ContentID headers, values are the attachments as Media Objects."}

    def __init__(self, msg_dict):
        self.envelope = Message.Envelope(msg_dict['envelope'])
        self.dkim = Message.DKIM(msg_dict['dkim'])
        self.addressees = Message.Addressees(msg_dict['addressees'])
        self.headers = msg_dict['headers']
        self.subject = msg_dict['subject']
        self.text = msg_dict['text']
        self.html = msg_dict['html']
        self.attachments = msg_dict['attachments']
        self.inline_attachments = msg_dict['inline_attachments']

    #!defMethod(_,header_name,[default=None])!2: "Return the value of the specified header, or default value if it is not present.\n\nCase-insensitive. If the header is specified multiple times, returns the first value." ["get_header"]
    def get_header(self, header_name, default=None):
        header_name = header_name.lower()
        for name,value in self.headers:
            if name.lower() == header_name:
                return value
        return default

    #!defMethod(_,header_name)!2: "Return a list containing every value of the specified header. Case-insensitive." ["list_header"]
    def list_header(self, header_name):
        header_name = header_name.lower()
        return [value for name,value in self.headers
                if name.lower() == header_name]

    #!defMethod(_,[cc=],[bcc=],[from_address=],[from_name=],[text=],[html=],[attachments=])!2: "Reply to this email." ["reply"]
    def reply(self,**kw):
        kw['to'] = kw.get('to', self.get_header("Reply-To", None))
        if kw['to'] is None:
            if self.addressees.from_address is not None:
                kw['to'] = self.addressees.from_address.raw_value
            else:
                kw['to'] = self.envelope.from_address
        if kw['to'] is None:
            raise Exception("Cannot reply to a message with no Reply-To header, From address, or Envelope From address.")

        kw['subject'] = kw.get('subject', self.subject)
        kw['in_reply_to'] = self.get_header("Message-ID")
        if kw['in_reply_to']:
            kw['references'] = self.get_header("References", "") + " " + kw['in_reply_to']
        kw['from_address'] = kw.get('from_address', self.envelope.recipient)
        send(**kw)

    def __str__(self):

        truncated_text = ""
        if self.text:
            truncated_text = self.text.replace("\n", " \\ ")
            (truncated_text[:70] + '...') if len(truncated_text) > 70 else truncated_text,

        return """anvil.email.Message:
    from: %s
    to: %s
    subject: %s
    text: %s
    attachments: %s""" % (
            self.addressees.from_address and self.addressees.from_address.raw_value,
            len(self.addressees.to_addresses) > 0 and self.addressees.to_addresses[0].raw_value,
            self.subject,
            truncated_text,
            ", ".join(["%s (%s bytes)" % (a.name, len(a.get_bytes())) for a in self.attachments]) if len(self.attachments) > 0 else None
        )

    #!defClass(anvil.email,#Message)!:


@anvil.server.portable_class
class SendReport(object):

    #!defAttr()!1: {name: "message_id", type: "string", description: "The Message-ID header given to this outgoing message."}

    def __init__(self):
        raise Exception("Cannot construct a SendReport manually")

    #!defClass(anvil.email,#SendReport)!:



import anvil.server
from anvil import *
from ..exceptions import AuthenticationFailed, MFAException
from ..config import get_client_config
from anvil.js import window

def replace(s, pattern, replacement):
    return window.String.prototype.replaceAll.call(s, window.RegExp(pattern, "g"), replacement)

class PhoneBox(TextBox):

    def __init__(self, **properties):
        TextBox.__init__(self, **properties)
        self.valid_number = None
        self.type="tel"
        self.add_event_handler("focus", self.on_focus)
        self.add_event_handler("lost_focus", self.on_blur)
        self.add_event_handler("pressed_enter", self.on_blur)

    def on_focus(self, **e):
        self.placeholder = "(123) 456 7890" if window.navigator.language == "en-US" else "+1 000 0..."

    def on_blur(self, **e):
        self.validate()

    def validate(self):
        leadingPlus = self.text.startswith("+")
        self.text = replace(self.text, "[^0-9]", "")
        if leadingPlus:
            if len(self.text) > 3: # What's a valid minimum length?
                self.text = "+" + self.text
                self.valid_number = self.text
            else:
                self.valid_number = None
        else:
            if len(self.text) == 10:
                self.valid_number = "+1" + self.text
                self.text = "(" + self.text[0:3] + ") " + self.text[3:6] + " " + self.text[6:]
            else:
                self.valid_number = None


#!defFunction(anvil.users,_,email_address)!2: "Send a two-factor authentication reset email to the specified user." ["send_mfa_reset_email"]
def send_mfa_reset_email(email):
    anvil.server.call("anvil.private.users.send_mfa_reset_email", email)
if is_server_side():
    pass
else:
    from . import webauthn

    #!defFunction(anvil.users.mfa,_,email_address)!2: "Generate a WebAuthn challenge that can be used to register a new hardware token for two-factor authentication." ["create_fido_mfa_method"]
    def create_fido_mfa_method(email):
        opts = anvil.server.call("anvil.private.users.begin_fido_attestation", email)
        result = webauthn.create(opts)
        if result:
            return anvil.server.call("anvil.private.users.validate_fido_attestation", result)

    #!defFunction(anvil.users.mfa,_,email_address,password)!2: "Generate a WebAuthn challenge that the given user can use to log in with a previously registered hardware token." ["get_fido_mfa_login"]
    def get_fido_mfa_login(email, password):
        opts = anvil.server.call("anvil.private.users.begin_fido_assertion", email, password)
        result = webauthn.get(opts)
        if result:
            return {"type": 'fido', "result": result}

    #!defFunction(anvil.users.mfa,_,code)!2: "Get an MFA login object representing a TOTP login code. This can be passed to the login_with_email function as the mfa argument." ["get_totp_mfa_login"]
    def get_totp_mfa_login(code):
        return {"type": "totp", "code": code}

    #!defFunction(anvil.users.mfa,_,code)!2: "Get an MFA login object representing a Twilio Verify token. This can be passed to the login_with_email function as the mfa argument." ["get_twilio_mfa_login"]
    def get_twilio_mfa_login(code):
        return {"type": "twilio-verify", "code": code}

    #!defFunction(anvil.users.mfa,_,email_address)!2: "Generate a TOTP secret that can be added as two-factor authentication for the current user." ["generate_totp_secret"]
    def generate_totp_secret(email):
        return anvil.server.call("anvil.private.users.totp.generate_secret", email)

    #!defFunction(anvil.users.mfa,_,mfa_method,code)!2: "Validate the given TOTP code against the given MFA method from a User row." ["validate_totp_code"]
    def validate_totp_code(mfa_method, code):
        return anvil.server.call("anvil.private.users.totp.validate_code", mfa_method, code)

    #!defFunction(anvil.users.mfa,_,phone)!2: "Generate a Twilio MFA method from the provided phone number." ["generate_twilio_mfa_method"]
    def generate_twilio_mfa_method(phone):
        return anvil.server.call("anvil.private.users.twilio.generate_mfa_method", phone)

    #!defFunction(anvil.users.mfa,_, mfa_method, channel)!2: "Send a Twilio Verify token using the given MFA method from a User row." ["send_twilio_token"]
    def send_twilio_token(mfa_method, channel):
        return anvil.server.call("anvil.private.users.twilio.send_verification_token", mfa_method, channel)

    #!defFunction(anvil.users.mfa,_, mfa_method, token)!2: "Validate the given Twilio Verify token against the given MFA method from a User row." ["check_twilio_token"]
    def check_twilio_token(mfa_method, token):
        return anvil.server.call("anvil.private.users.twilio.check_verification_token", mfa_method, token)

    #!defFunction(anvil.users.mfa,_,password,mfa_method,[clear_existing=False])!2: "Add an MFA method to the current user by passing the user's password and the mfa method, optionally clearing all existing methods." ["add_mfa_method"]
    def add_mfa_method(password, method, clear_existing=False):
        return anvil.server.call("anvil.private.users.add_mfa_method", password, method, clear_existing)

    #!defFunction(anvil.users.mfa,_,email_address,password)!2: "Get the available MFA types for the given user by passing their email and password." ["get_available_mfa_types"]
    def get_available_mfa_types(email, password):
        return anvil.server.call("anvil.private.users.get_available_mfa_types", email, password)

    #!defFunction(anvil.users.mfa,_)!2: "Get all the enabled MFA types for this app." ["get_enabled_mfa_types"]
    def get_enabled_mfa_types():
        return anvil.server.call("anvil.private.users.get_enabled_mfa_types")

    def _configure_mfa(email, mfa_error, require_password, allow_cancel, confirm_button_text):
        mfa_types = get_enabled_mfa_types()
        selected_mfa_type = None
        mfa_methods = {}
        password_error = None
        password_box = TextBox(placeholder="Current password", align="center", hide_text=True)

        while True:
            mfa_panel = LinearPanel()

            if require_password and not password_box.text:
                password_box.remove_from_parent()
                mfa_panel.add_component(password_box)
                if password_error:
                    mfa_panel.add_component(Label(foreground="red", align="center", spacing_below="none", text=password_error))

            if not selected_mfa_type:

                mfa_panel.add_component(Label(text="This app requires 2-factor authentication to log in."))


                if "fido" in mfa_types:
                    def signup_with_fido(**e):
                        result = create_fido_mfa_method(email)
                        if result:
                            mfa_methods['fido'] = result
                            mfa_panel.raise_event("x-close-alert", value='fido')

                    if webauthn.is_webauthn_available():
                        fido_link = Link(text="Use hardware token",icon="fa:lock")
                        fido_link.set_event_handler("click", signup_with_fido)
                        mfa_panel.add_component(fido_link)
                    else:
                        mfa_panel.add_component(Label(text="Hardware token unavailable", icon="fa:lock", tooltip="Is this page running in an iframe or an unsupported browser?"))

                if "totp" in mfa_types:
                    totp_link = Link(text="Use Authenticator app", icon="fa:qrcode")
                    totp_link.set_event_handler("click", lambda **e: mfa_panel.raise_event("x-close-alert", value='select-totp'))
                    mfa_panel.add_component(totp_link)

                if "twilio-verify" in mfa_types:
                    twilio_link = Link(text="Use phone number", icon="fa:phone")
                    twilio_link.set_event_handler("click", lambda **e: mfa_panel.raise_event("x-close-alert", value='select-twilio'))
                    mfa_panel.add_component(twilio_link)

            else:
                back_link = Link(text="Choose another method", icon="fa:arrow-left")
                back_link.set_event_handler("click", lambda **e: mfa_panel.raise_event("x-close-alert", value='back'))
                mfa_panel.add_component(back_link)

                if selected_mfa_type == "totp":
                    totp_config = generate_totp_secret(email)
                    totp_secret = totp_config['secret']
                    qr_code = totp_config['qr_code']
                    mfa_methods['totp'] = totp_config['mfa_method']

                    mfa_panel.add_component(Label(text="Scan this QR-code with your Authenticator app and then enter the current code below to continue."))
                    mfa_panel.add_component(Image(source=qr_code, display_mode="fill_width"))

                    totp_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
                    totp_box.set_event_handler("show", lambda **e: totp_box.focus())
                    totp_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event("x-close-alert"))
                    mfa_panel.add_component(totp_box)

                elif selected_mfa_type == "twilio-verify":
                    if not mfa_methods.get('twilio-verify'):
                        phone_box = PhoneBox(placeholder="Enter your phone number", align="center")
                        phone_box.add_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event("x-close-alert"))
                        #phone_box.add_event_handler("lost_focus", lambda **e: print(phone_box.valid_number))
                        #mfa_panel.add_component(Label(text="Enter your phone number", align="center"))
                        mfa_panel.add_component(phone_box)
                    else:
                        twilio_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
                        twilio_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event("x-close-alert"))
                        twilio_box.set_event_handler("show", lambda **e: twilio_box.focus())
                        mfa_panel.add_component(twilio_box)                            

                        sms_link = Link(text="Resend Text Message", icon="fa:commenting-o")
                        phone_link = Link(text="Call Me instead", icon="fa:phone")
                        sms_link.set_event_handler("click", lambda **e: mfa_panel.raise_event("x-close-alert", value='resend-sms'))
                        phone_link.set_event_handler("click", lambda **e: mfa_panel.raise_event("x-close-alert", value='call'))
                        mfa_panel.add_component(sms_link)
                        mfa_panel.add_component(phone_link)

            if mfa_error:
                mfa_panel.add_component(Label(foreground="red", bold=True, spacing_below="none", text=mfa_error))

            maybe_cancel_button = [("Cancel", 'cancel')] if allow_cancel else []
            m = alert(mfa_panel, title="2-Factor Authentication", buttons=[(confirm_button_text, True, 'success')] + maybe_cancel_button, dismissible=bool(allow_cancel))
            if m == 'cancel':
                return None, ""
            elif require_password and not password_box.text: # TODO: Also validate password
                password_error = "Please enter your password"
            elif m == 'back':
                selected_mfa_type = None
                mfa_error = None
            elif m == 'select-totp':
                mfa_error = None
                mfa_methods['totp'] = None
                selected_mfa_type = 'totp'
            elif m == 'select-twilio':
                mfa_error = None
                mfa_methods['twilio-verify'] = None
                selected_mfa_type = 'twilio-verify'
            elif m == 'fido':
                # fido is a single step process - we're done
                return mfa_methods['fido'], password_box.text

            elif selected_mfa_type == 'totp':
                if validate_totp_code(mfa_methods['totp'], totp_box.text):
                    return mfa_methods['totp'], password_box.text
                else:
                    mfa_error = "Incorrect code entered. Please try again."
            elif selected_mfa_type == 'twilio-verify':
                mfa_error = None
                if m == 'resend-sms':
                    channel = "sms"
                    mfa_methods['twilio-verify'] = None
                    mfa_error = "Text message resent"
                elif m == 'call':
                    channel = "call"
                    mfa_methods['twilio-verify'] = None
                    mfa_error = "Calling you now"
                else:
                    channel = "sms"

                if not mfa_methods.get('twilio-verify'):
                    if phone_box.valid_number:
                        mfa_methods['twilio-verify'] = generate_twilio_mfa_method(phone_box.valid_number)
                        try:
                            send_twilio_token(mfa_methods['twilio-verify'], channel)
                        except MFAException as e:
                            mfa_error = e.message
                    else:
                        mfa_error = "Please enter a valid phone number."
                else:
                    if check_twilio_token(mfa_methods['twilio-verify'], twilio_box.text):
                        return mfa_methods['twilio-verify'], password_box.text
                    else:
                        mfa_error = "Incorrect code entered. Please try again."
            else:
                mfa_error = None


    #!defFunction(anvil.users.mfa,!,email_address, password)!2: "Display a form to collect two-factor authentication credentials from the user currently logging in by passing the function their email and password." ["mfa_login_with_form"]
    def mfa_login_with_form(email, password):
        mfa_panel = LinearPanel()

        mfa_types = get_available_mfa_types(email, password)

        maybe_login_button = []

        error_label = Label(foreground="red", bold=True, spacing_below="none",visible=False)
        mfa_panel.add_component(error_label)
        def error(text):
            error_label.text = text
            error_label.visible = True

        def clear_error():
            error_label.visible = False
            error_label.text = ""
        
        if 'totp' in mfa_types:
            totp_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
            totp_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event('x-close-alert', value=get_totp_mfa_login(totp_box.text)))
            totp_box.set_event_handler("show", lambda **e: totp_box.focus())
            mfa_panel.add_component(Label(text="Please enter the 2-factor authentication code from your Authenticator app"))
            mfa_panel.add_component(totp_box)
            maybe_login_button = [("Log In", 'totp', 'success')]

        if 'fido' in mfa_types:
            def use_fido(**e):
                val = get_fido_mfa_login(email, password)
                if val:
                    mfa_panel.raise_event("x-close-alert", value=val)

            
            if webauthn.is_webauthn_available():
                fido_link = Link(text="Use hardware token", icon="fa:lock")
                fido_link.set_event_handler("click", use_fido)
                mfa_panel.add_component(fido_link)
            else:
                mfa_panel.add_component(Label(text="Hardware token unavailable", icon="fa:lock", tooltip="Is this page running in an iframe or an unsupported browser?"))

        if 'twilio-verify' in mfa_types:
            
            def use_phone(**e):
                error("Calling you now")
                try:
                    send_twilio_token(None, "call")
                except MFAException as e:
                    error(e.message)
            def use_sms(display_sent=True, **e):
                if display_sent:
                    error("Text message resent")
                try:
                    send_twilio_token(None, "sms")
                except MFAException as e:
                    error(e.message)

            use_sms(False)

            twilio_box = TextBox(placeholder="Enter 6-digit code", align="center", font="monospace")
            twilio_box.set_event_handler("pressed_enter", lambda **e: mfa_panel.raise_event('x-close-alert', value=get_twilio_mfa_login(twilio_box.text)))
            twilio_box.set_event_handler("show", lambda **e: twilio_box.focus())
            mfa_panel.add_component(Label(text="Please enter the 6-digit code we sent to the phone number we have on file"))
            mfa_panel.add_component(twilio_box)

            sms_link = Link(text="Resend Text Message", icon="fa:commenting-o")
            phone_link = Link(text="Call Me instead", icon="fa:phone")
            sms_link.set_event_handler("click", use_sms)
            phone_link.set_event_handler("click", use_phone)
            mfa_panel.add_component(sms_link)
            mfa_panel.add_component(phone_link)

            maybe_login_button = [("Log In", 'twilio', 'success')]



        if not mfa_types:
            mfa_panel.add_component(Label(text="No authentication methods available."))

        if not mfa_types or get_client_config().get("allow_mfa_email_reset", False):
            mfa_reset_link = Link(text="Reset 2-factor authentication by email")
            mfa_reset_link.set_event_handler('click', lambda **e: mfa_panel.raise_event('x-close-alert', value='reset_mfa'))
            mfa_panel.add_component(mfa_reset_link)

        r = alert(mfa_panel, title="2-Factor Authentication", buttons=maybe_login_button + [("Cancel", None, 'default')], dismissible=False)

        if r == 'totp':
            return get_totp_mfa_login(totp_box.text)
        elif r == 'twilio':
            return get_twilio_mfa_login(twilio_box.text)
        else:
            return r


    #!defFunction(anvil.users.mfa,!,[allow_cancel=False])!2: "Display a form for the user to configure 2-factor authentication.\n\nallow_cancel: if True, the signup form has a Cancel button that the user can use to dismiss the form." ["configure_mfa_with_form"]
    def configure_mfa_with_form(allow_cancel=False):

        error = None
        while True:
            mfa_method, password = _configure_mfa(None, error, True, allow_cancel, "Save")

            if mfa_method:
                try:
                    add_mfa_method(password, mfa_method)
                    alert("Your two-factor authentication configuration has been reset.")
                    return True
                except AuthenticationFailed as e:
                    error = e.args[0]
                except Exception as e:
                    error = str(e)
            else:
                return None



import random
import os
import sys
import anvil
import tempfile
import io

open_ = open

class TempFile():

    #!defMethod(string,[media])!2: "Create a temporary file initialised with the contents of the provided media, if any." ["__init__"]
    def __init__(self, media=None):
        self._media = media

    #!defMethod(string)!2: "" ["__enter__"]
    def __enter__(self):
        self._filename = tempfile.gettempdir() + os.sep + "".join([random.choice("1234567890abcdefghijklmnopqrstuvwxyz") for i in range(32)])
        if self._media is not None:
            with open_(self._filename, "wb") as f:
                f.write(self._media.get_bytes())
        return self._filename

    #!defMethod(_)!2: "" ["__exit__"]
    def __exit__(self, e_type, e_val, tb):
        try:
            os.unlink(self._filename)
        except:
            # If it's already gone, we don't care.
            pass
#!defClass(anvil.media,%TempFile)!:


#!defFunction(anvil.media,%anvil.Media instance,filename,[mime_type],[name])!2: "Creates a Media object from the given file." ["from_file"]
def from_file(filename, mime_type=None, name=None):
    with open_(filename, "rb") as f:
        return anvil.BlobMedia(mime_type, f.read(), name=(name or filename.split(os.sep)[-1]))

#!defFunction(anvil.media,_,media,filename)!2: "Write a Media object to the given file" ["write_to_file"]
def write_to_file(media, filename):
    with open_(filename, "wb") as f:
        f.write(media.get_bytes())


#!defFunction(anvil.media,%BytesIO, media)!2: "Open a media file as Python BytesIO object" ["open"]
def open(media):
    return io.BytesIO(media.get_bytes())


import plotly.graph_objs as go

try:
    _TemplateFactory = go.layout.Template
except AttributeError:
    _TemplateFactory = dict


_cache = {}

def _get_pio():
    try:
        import plotly.io as pio
    except ImportError:
        return None
    else:
        return pio


def _register_template(name, defn):
    pio = _get_pio()
    if pio is None:
        return
    
    if name not in pio.templates:
        pio.templates[name] = _TemplateFactory(**defn)



class Templates:
    def __getitem__(self, theme):
        try:
            return _cache[theme]
        except KeyError:
            pass

        import os
        import json

        current_dir = os.path.dirname(os.path.realpath(__file__))
        file_path = os.path.join(current_dir, "templates", theme + ".json")

        try:
            with open(file_path, "r") as file:
                data = json.load(file)
        except Exception:
            raise KeyError(theme)
        else:
            _cache[theme] = data
            _register_template(theme, data)

        return data

    def __getattr__(self, theme):
        try:
            return self[theme]
        except KeyError:
            raise AttributeError(theme)


#!defModuleAttr(anvil.plotly_templates)!1: {name: "templates", type: "dict", description: "A dictionary of custom templates for anvil themes"}
templates = Templates()


#!defFunction(anvil.plotly_templates,_,template)!2: "Sets the default plotly template" ["set_default"]
def set_default(template):
    pio = _get_pio()

    if pio is None:
        go.Figure._default_template = template
    elif type(template) is str:
        templates[template]
        pio.templates.default = template
    else:
        pio.templates.default = template



def _get_template(self, template):
    return templates[template]


go.Figure._get_template = _get_template



from .._base_classes import Row, SearchIterator, Table
from . import _load_hacks
from ._app_tables import app_tables, get_table_by_id

# from ._batcher import batch_delete, batch_update

__all__ = ["app_tables", "get_table_by_id"]



import anvil.server

#def charge(token, amount, currency):
#    return anvil.server.call("anvil.private.stripe.charge", amount, currency)

def subscribe_new_customer(token, email_address, plan, quantity=1):
    customer = anvil.server.call("anvil.private.stripe.new_customer", email_address, token)
    customer.subscribe(plan, quantity)
    return customer

#!defFunction(anvil.stripe,stripe..Customer instance,email_address,[token])!2: "Create a new Stripe Customer record" ["new_customer"]
def new_customer(email_address, token=None):
    return anvil.server.call("anvil.private.stripe.new_customer", email_address, token)

#!defFunction(anvil.stripe,stripe..Customer instance,customer_id)!2: "Retrieve a Stripe Customer record by its ID" ["get_customer"]
def get_customer(customer_id):
    return anvil.server.call("anvil.private.stripe.get_customer", customer_id)


# Virtual definitions of live-object types
#!defMethod(list[stripe..Subscription instance],[live_only])!2: "List all current subscriptions for this user.\n\nBy default, only returns live subscriptions" ["get_subscriptions"]
#!defMethod(list[str instance],[live_only])!2: "List the string ID for each current subscriptions for this user" ["get_subscription_ids"]
#!defMethod(stripe..Subscription instance,plan_id,[quantity])!2: "Subscribe this user to the specified plan (quantity defaults to 1)" ["new_subscription"]
#!defMethod(_,token)!2: "Add the specified token as a payment method for this customer" ["add_token"]
#!defMethod(_,amount,currency)!2: "Issue a one-off charge to this customer" ["charge"]
#!defClass(stripe.,Customer)!:


#!defMethod(_)!2: "Returns True if this subscription's status is 'active', 'trialling', or 'past_due'" ["is_live"]
#!defMethod(_,[at_period_end=True])!2: "Cancel this subscription.\nBy default, at_period_end=True, meaning the subscription will be cancelled at the end of the current billing period, with no credit or refund." ["cancel"]
#!defMethod(_,plan_id)!2: "Set this subscription to the specified Stripe plan" ["set_plan"]
#!defClass(stripe.,Subscription)!:




import anvil.server
from anvil.server import Capability

from .._base_classes import Table as BaseTable
from ._constants import CASCADE, KNOWN_PERMS, READ, SERVER_PREFIX, WRITE
from ._refs import make_refs
from ._row import Row
from ._search import SearchIterator
from ._utils import validate_cap

PREFIX = SERVER_PREFIX + "table."


@anvil.server.portable_class
class Table(BaseTable):
    @classmethod
    def _create(cls, cap, view_key, table_id):
        assert cap is None or type(cap) is Capability, "expected a table capability"
        self = object.__new__(cls)
        self._cap = cap
        self._view_key = view_key
        self._id = str(table_id)
        return self

    @classmethod
    def __new_deserialized__(cls, data, info):
        if not info.remote_is_trusted:
            validate_cap(*data)
        return cls._create(*data)

    def __serialize__(self, _info):
        return [self._cap, self._view_key, self._id]

    def __iter__(self):
        raise TypeError("You can't iterate on a table. Call search() on this table to get an iterator of rows instead.")

    def __eq__(self, other):
        if not isinstance(other, Table):
            return NotImplemented
        return other._id == self._id

    def __hash__(self):
        return hash(self._id)

    def __contains__(self, row):
        return self.has_row(row)

    def _get_view(self, perm, args, kws):
        assert perm in KNOWN_PERMS, "bad permission"
        new_cap, view_key = anvil.server.call(
            PREFIX + "get_view", self._cap, perm, None, make_refs(args), make_refs(kws)
        )
        return Table._create(new_cap, view_key, self._id)

    # PUBLIC API
    def restrict_columns(self, col_spec):
        new_cap, view_key = anvil.server.call("get_restricted_columns", self._cap, col_spec)
        return Table._create(new_cap, view_key, self._id)

    def client_readable(self, *args, **kws):
        return self._get_view(READ, args, kws)

    def client_writable(self, *args, **kws):
        return self._get_view(WRITE, args, kws)

    def client_writable_cascade(self, *args, **kws):
        return self._get_view(CASCADE, args, kws)

    def delete_all_rows(self):
        return anvil.server.call(PREFIX + "delete_all_rows", self._cap)

    def add_rows(self, rows):
        # rows can be an iterable of dicts
        row_dicts = []
        refs = []
        for row in rows:
            row = dict(row)
            refs.append(make_refs(row))
            row_dicts.append(row)
        row_id_caps, spec = anvil.server.call(PREFIX + "add_rows", self._cap, refs)
        return [
            Row._create_from_local_values(self._view_key, self._id, row_id, spec, cap, row_items)
            for (row_id, cap), row_items in zip(row_id_caps, row_dicts)
        ]

    def add_row(self, **data):
        row_id, cap, spec = anvil.server.call(PREFIX + "add_row", self._cap, make_refs(data))
        return Row._create_from_local_values(self._view_key, self._id, row_id, spec, cap, data)

    def get(self, *args, **kws):
        row_id_table_data = anvil.server.call(PREFIX + "get_row", self._cap, make_refs(args), make_refs(kws))
        return row_id_table_data and Row._create_from_trusted(self._view_key, self._id, *row_id_table_data)

    def get_by_id(self, row_id, fetch=None):
        row_id_table_data = anvil.server.call(PREFIX + "get_row_by_id", self._cap, row_id, fetch=fetch)
        return row_id_table_data and Row._create_from_trusted(self._view_key, self._id, *row_id_table_data)

    def has_row(self, row):
        if not isinstance(row, Row):
            # backwards compatability return False
            return False
        elif row._table_id != self._id:
            return False
        return anvil.server.call(PREFIX + "has_row", self._cap, row._id)

    def list_columns(self):
        return anvil.server.call(PREFIX + "list_columns", self._cap)

    def search(self, *args, **kws):
        kws = make_refs(kws)
        row_ids, cap, cap_next, table_data = anvil.server.call(PREFIX + "search", self._cap, args, kws)
        return SearchIterator._create(self._view_key, self._id, row_ids, cap, cap_next, table_data)

    def to_csv(self, escape_for_excel=False):
        return anvil.server.call(PREFIX + "to_csv", self._cap, escape_for_excel=escape_for_excel)

    # TODO reinclude this API
    # @property
    # def id(self):
    #     return self._id



import anvil.server

def login(*a, **k):
    raise Exception("anvil.microsoft.auth.login cannot be called from a server module")

def get_user_email():
    return anvil.server.call("anvil.private.microsoft.auth.get_user_email")

def get_user_id():
    return anvil.server.call("anvil.private.microsoft.auth.get_user_id")

def get_user_access_token():
    return anvil.server.call("anvil.private.microsoft.auth.get_user_access_token")

def get_user_refresh_token():
    return anvil.server.call("anvil.private.microsoft.auth.get_user_refresh_token")

def refresh_access_token(refresh_token):
    return anvil.server.call("anvil.private.microsoft.auth.refresh_access_token", refresh_token)



from sqlalchemy import MetaData, Engine, select, insert, update, delete
from . import DatabaseImpl, Table, TableInfo, Row
import anvil.server
from anvil.server import unwrap_capability, Capability
from datetime import date, datetime


class Database(DatabaseImpl):
    def __init__(self, db_name: str, engine: Engine):
        self.db_name = db_name
        self.engine = engine
        self.md = MetaData()
        self.md.reflect(bind=engine)
        self.table_info = {table.name: ConcreteTableInfo(self, table)
                           for table in self.md.tables.values()}
        # In future, we will do clever things like detecting foreign key relationships

    def search(self, table: TableInfo):
        with self.engine.connect() as conn:
            real_table = self.md.tables[table.name]
            result = conn.execute(select(real_table))
            colnames = [col.key for col in real_table.columns]
            return [Row(table=table, data={name: row[name] for name in colnames}) for row in result.mappings()]

    def add_row(self, table: TableInfo, data: dict):
        with self.engine.connect() as conn:
            real_table = self.md.tables[table.name]
            result = conn.execute(insert(real_table).values(**data))

            # Fill in primary key
            for col, value in zip(table.primary_key, result.inserted_primary_key):
                data[col] = value

            conn.commit()

            return Row(data=data, table=table)

    def update_row(self, table: TableInfo, primary_key: tuple, updates: dict):
        if not primary_key:
            raise TypeError(f"The \"{table.name}\" table has no PRIMARY KEY and therefore cannot be updated")
        if any(col in updates for col in table.primary_key):
            raise ValueError("Cannot change the primary key of a row")
        with self.engine.connect() as conn:
            real_table = self.md.tables[table.name]
            conn.execute(update(real_table).where(*(col == val for col,val in zip(real_table.primary_key.columns, primary_key))).values(updates))
            conn.commit()

    def delete_row(self, table: TableInfo, primary_key: tuple):
        if not primary_key:
            raise TypeError(f"The \"{table.name}\" table has no PRIMARY KEY and therefore cannot be updated")
        with self.engine.connect() as conn:
            real_table = self.md.tables[table.name]
            conn.execute(delete(real_table).where(*(col == val for col,val in zip(real_table.primary_key.columns, primary_key))))
            conn.commit()


class ConcreteTableInfo(TableInfo):
    def __init__(self, db: Database, real_table):
        pk = tuple(col.name for col in real_table.primary_key.columns)
        columns = {col.name: {} for col in real_table.columns}
        super().__init__(db_name=db.db_name, name=real_table.name, primary_key=pk, columns=columns, db=db)

    @property
    def safe_version(self):
        return TableInfo(self.db.db_name, self.name, self.primary_key, self.columns, None)


class ExternalTablesImpl:
    def __init__(self, db_name, engine):
        self.db = Database(db_name, engine)
        setup_callables(self.db)

    def get_table(self, name):
        if name in self.db.table_info:
            return Table(info=self.db.table_info[name])


class ExternalTables:
    # The entry point to this API. Initialise with an SQLAlchemy Engine object, then it behaves
    # like anvil.tables.app_tables.
    def __init__(self, name, engine):
        et_impl = ExternalTablesImpl(name, engine)
        object.__setattr__(self, "impl", et_impl)

    def __getattribute__(self, name):
        tbl = object.__getattribute__(self, "impl").get_table(name)
        if tbl is None:
            raise AttributeError(name)
        return tbl

    def __setattr__(self, key, value):
        raise TypeError("This object is read-only")


KNOWN_COLUMN_TYPES = {
    str: "string",
    bool: "bool",
    date: "date",
    datetime: "datetime",
    int: "number",
    float: "number",
}

def setup_callables(db: Database):
    @anvil.server.callable("anvil.ext_tables.search/"+db.db_name)
    def search(cap):
        _, table_name = anvil.server.unwrap_capability(cap, ["ext_tables/"+db.db_name, Capability.ANY])
        return db.search(db.table_info[table_name])

    @anvil.server.callable("anvil.ext_tables.add_row/"+db.db_name)
    def add_row(cap, data):
        _, table_name = anvil.server.unwrap_capability(cap, ["ext_tables/"+db.db_name, Capability.ANY])
        return db.add_row(db.table_info[table_name], data)

    @anvil.server.callable("anvil.ext_tables.update_row/"+db.db_name)
    def update_row(cap, updates):
        _, table_name, primary_key = anvil.server.unwrap_capability(cap, ["ext_tables/"+db.db_name, Capability.ANY, Capability.ANY])
        db.update_row(db.table_info[table_name], primary_key, updates)

    @anvil.server.callable("anvil.ext_tables.delete_row/"+db.db_name)
    def delete_row(cap):
        _, table_name, primary_key = anvil.server.unwrap_capability(cap, ["ext_tables/"+db.db_name, Capability.ANY, Capability.ANY])
        db.delete_row(db.table_info[table_name], primary_key)

    @anvil.server.callable("anvil.record_schema.get/anvil.ext_tables."+db.db_name)
    def get_schema():
        return {
            "name": "An SQL database",
            "record_types": [
                {"id": name, "name": name, "fields": [
                    {"key": col.key, "type": KNOWN_COLUMN_TYPES[col.type.python_type]}
                    for col in table.columns
                    if col.type.python_type in KNOWN_COLUMN_TYPES
                ]}
                for name, table in db.md.tables.items()
            ]
        }


from anvil.server import serializable_type


def _wrap(value):
    if isinstance(value, (WrappedObject, WrappedList)):
        return value
    elif isinstance(value, dict):
        return WrappedObject(value)
    elif isinstance(value, list):
        wl = WrappedList()
        for i in value:
            wl.append(i)
        return wl
    else:
        return value


@serializable_type
class WrappedObject(dict):
    _name = None
    _module = None

    def __init__(self, d=None, **kwargs):
        if d and isinstance(d, dict):
            for k in d.keys():
                self.__setitem__(k, d[k])

        for k in kwargs.keys():
            self.__setitem__(k, kwargs[k])

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, _wrap(value))

    def __getitem__(self, key):
        _sentinel = WrappedObject()
        r = dict.get(self, key, _sentinel)

        if r is _sentinel:
            dict.__setitem__(self, key, _sentinel)

        return r

    def update(*args, **new_items):
        # avoid name conflicts with keys, could use (self, other=(), /, **kws)
        # but position only args not available in py2/Skulpt
        if not args:
            raise TypeError("method 'update' needs an argument")
        elif len(args) > 2:
            raise TypeError("expected at most 1 argument, got %d" % (len(args) - 1))
        elif len(args) == 2:
            new_items = dict(args[1], **new_items)
        self = args[0]
        for k, v in new_items.items():
            self[k] = v

    def __repr__(self):
        n = self._name or "WrappedObject"
        m = self._module + "." if self._module else ""
        return "%s%s<%s>" % (m, n, ", ".join(["%s=%s" % (k, repr(self[k])) for k in self.keys()]))

    def __serialize__(self, global_data):
        return dict(self)

    def __deserialize__(self, data, global_data):
        self.__init__(data)

    def __copy__(self):
        return self.__class__(dict.copy(self))

    def __deepcopy__(self, memo):
        # lazy load this - its only need on the
        # server and we don't want to load copy on the client
        from copy import deepcopy

        return self.__class__(deepcopy(dict(self)))


@serializable_type
class WrappedList(list):
    def __init__(self, lst=[]):
        for x in lst:
            self.append(x)

    def append(self, item):
        list.append(self, _wrap(item))

    def extend(self, items):
        for i in items:
            self.append(i)

    def insert(self, offset, item):
        list.insert(self, offset, _wrap(item))

    def __serialize__(self, global_data):
        return list(self)

    def __deserialize__(self, data, global_data):
        self.__init__(data)

    def __copy__(self):
        return self.__class__(list.copy(self))

    def __deepcopy__(self, memo):
        from copy import deepcopy

        return self.__class__(deepcopy(list(self)))



# Implementation of anvil.server for the downlink worker

from ._threaded_server import live_object_backend, LazyMedia, call_info as session, _switch_session, call_context as context

from ._server import (register, 
                      callable, 
                      background_task, 
                      callable_as, 
                      Serializable,
                      portable_class,
                      serializable_type,
                      Capability,
                      unwrap_capability,
                      _register_exception_type, 
                      AnvilWrappedError, 
                      SerializationError, 
                      InternalError, 
                      InvalidResponseError, 
                      RuntimeUnavailableError, 
                      QuotaExceededError, 
                      UplinkDisconnectedError, 
                      ExecutionTerminatedError, 
                      TimeoutError, 
                      NoServerFunctionError, 
                      PermissionDenied,
                      ServiceNotAdded,
                      CookieError, 
                      _FailError,
                      BackgroundTaskError,
                      BackgroundTaskNotFound,
                      BackgroundTaskKilled,
                      http_endpoint,
                      wellknown_endpoint,
                      api_request as request, 
                      HttpResponse, 
                      cookies,
                      raise_event,
                      list_client_sessions,
                      get_client_session,
                      get_session_id,
                      subscribe,
                      unsubscribe,
                      get_subscriptions)

from . import _threaded_server, _server

#!defFunction(anvil.server,%, function_name, **kwargs)!2: "Allows you to call server-side functions from the client. It requires the name of the function that you want to call, passed as a string. Any extra arguments are passed through to the server function." ["call"]
def call(*args, **kwargs):
    if not args:
        raise TypeError("anvil.server.call() expects atleast 1 argument")
    fn_name, args = args[0], args[1:]
    if not isinstance(fn_name, str):
        raise TypeError("first argument to anvil.server.call() must be as str, got '" + type(fn_name).__name__ + "'")
    try:
        return _threaded_server.do_call(args, kwargs, fn_name=fn_name)
    except _server.AnvilWrappedError as e:
        error_from_server = _server._deserialise_exception(e.error_obj)
        raise error_from_server


# Once downlinks rebuilt and updated for all of legacy, dynamic and sandbox, update these to include [prefer_ephemeral_debug=] and update docs
#!defFunction(anvil.server,string,[environment_type])!2: {anvil$args: {environment: "Pass 'published' to get the published URL"}, anvil$helpLink: "/docs/http-apis/creating-http-endpoints#getting-the-url-for-your-api", $doc: "Returns the root URL for the current app.\n\nBy default, this function returns the URL for the current environment, which might be private or temporary (for example, if you are running your app in the Anvil Editor). If you want the URL for the published branch, pass 'published' as an argument."} ["get_app_origin"]
def get_app_origin(environment_type=None, **kwargs):
    return call("anvil.private.get_app_origin", environment_type, **kwargs)


#!defFunction(anvil.server,string,[environment_type])!2: {anvil$args: {environment: "Pass 'published' to get the published URL"}, anvil$helpLink: "/docs/http-apis/creating-http-endpoints#getting-the-url-for-your-api", $doc: "Returns the root URL of the API for the current app.\n\nBy default, this function returns the URL for the current environment, which might be private or temporary (for example, if you are running your app in the Anvil Editor). If you want the URL for the published branch, pass 'published' as an argument."} ["get_api_origin"]
def get_api_origin(environment_type=None, **kwargs):
    return call("anvil.private.get_api_origin", environment_type, **kwargs)


def launch_background_task(fn_name, *args, **kwargs):
    return call("anvil.private.background_tasks.launch", fn_name, *args, **kwargs)


def get_background_task(id):
    return call("anvil.private.background_tasks.get_by_id", id)


def list_background_tasks(all_environments=False):
    return call("anvil.private.background_tasks.list", all_environments=all_environments)


task_state = _server.NotABackgroundTaskState()

#!defAttr()!1: {name: "headers", type: "dict", description: "HTTP headers sent with the current request"}
#!defAttr()!1: {name: "method", type: "string", description: "The HTTP method of the current request, e.g. GET, POST, etc."}
#!defAttr()!1: {name: "path", type: "string", description: "The path of the current request, e.g. /foo/bar"}
#!defAttr()!1: {name: "origin", type: "string", description: "The origin of the current API request, e.g. https://my-app.anvil.app/_/api"}
#!defAttr()!1: {name: "remote_address", type: "string", description: "The IP address the current request is coming from."}
#!defAttr()!1: {name: "body", pyType: "anvil.Media instance", description: "The body of the HTTP request"}
#!defAttr()!1: {name: "body_json", type: "dict", description: "The decoded JSON body of the HTTP request, if applicable. Only available when Content-Type header is 'application/json'."}
#!defAttr()!1: {name: "username", type: "string", description: "The username received through HTTP Basic Authentication"}
#!defAttr()!1: {name: "password", type: "string", description: "The password received through HTTP Basic Authentication"}
#!defAttr()!1: {name: "user", type: "User", description: "When require_auth is True, returns the row from the Users table corresponding to the authenticated user."}
#!defAttr()!1: {name: "query_params", type: "dict", description: "A dict of query-string parameters passed with this request."}
#!defAttr()!1: {name: "form_params", type: "dict", description: "A dict of form parameters passed with this request."}

#!defClass(anvil.server,%HttpRequest)!0:

#!defModuleAttr(anvil.server)!1: {name: "%request", pyType: "anvil.server.HttpRequest instance", description: "Contains information about the current HTTP API request. It's an instance of `anvil.server.HttpRequest`."}

#!defAttr()!1: {name: "status", type: "number", description: "The status code for this HTTP response. Default is 200."}
#!defAttr()!1: {name: "body", type: "any", description: "The body of this HTTP response. Can be a string, a Media object, or any JSON-able value."}
#!defAttr()!1: {name: "headers", type: "dict", description: "The headers to return with this HTTP response. Content-Type will be set automatically if not specified."}
#!defMethod(_, status=200, body="", headers=None)!2: ("Create an HttpResponse object") ["__init__"];
#!defClass(anvil.server,%HttpResponse)!0:


#!defAttr()!1: {name: "client", pyType: "anvil.server.CallContext.ClientInfo instance", description: "An object that describes the client that initiated the current session. This can be a browser, an HTTP endpoint request, an uplink script, a background task, or an incoming email."}
#!defAttr()!1: {name: "type", type: "string", description: "The execution environment this code is running in. May be 'browser', 'server_module' or 'uplink'"}
#!defAttr()!1: {name: "remote_caller", pyType: "anvil.server.CallContext.StackFrame instance", description: "An object describing the code that called this @anvil.server.callable function, where it was running, and whether it is trusted (server-side) or un-trusted (input from a browser, HTTP or other remote code)"}
#!defAttr()!1: {name: "background_task_id", type: "string", description: "The ID of the currently running background task, if there is one."}
#!defClass(anvil.server,#CallContext)!0:

#!defAttr()!1: {name: "ip", type: "string", description: "The IP address of the client that initiated this session."}
#!defAttr()!1: {name: "location", pyType: "anvil.server.CallContext.Location instance", description: "The location of this client, as determined by its IP address, or None if it cannot be determined."}
#!defAttr()!1: {name: "type", type: "string", description: "How this session was initiated. Valid values are: 'browser', 'uplink', 'http', 'background_task' and 'email'."}
#!defClass(anvil.server,#CallContext.ClientInfo)!0:


#!defAttr()!1: {name: "type", type: "string", description: "The location of the calling code. Valid values are: 'browser', 'server_module', 'uplink' and 'client_uplink' for your app's code, or 'http', 'background_task' or 'email' if this code was extrernally triggered."}
#!defAttr()!1: {name: "is_trusted", type: "boolean", description: "Was this code running in a trusted location (ie on the server side)?"}
#!defClass(anvil.server,#CallContext.StackFrame)!0:


#!defAttr()!1: {name: "latitude", type: "float"}
#!defAttr()!1: {name: "longitude", type: "float"}
#!defAttr()!1: {name: "city", type: "string"}
#!defAttr()!1: {name: "subdivision", type: "string"}
#!defAttr()!1: {name: "country", type: "string"}
#!defClass(anvil.server,#CallContext.Location)!0:

#!defModuleAttr(anvil.server)!1: {name: "context", pyType: "anvil.server.CallContext instance", description: "Contains information about what triggered the currently running code. It's an instance of `anvil.server.CallContext`."}

#!defModuleAttr(anvil.server)!1: {name: "session", pyType: "dict instance", description: "A dictionary that contains information about a specific session. It can store anything that a server function can return, except for Media objects."}

#!defFunction(anvil.server [uplink],#,key,[init_session=None],[quiet=False])!2: {$doc: "Connect your uplink script to your anvil app.", anvil$args: {keys: "The key is a unique string and should be kept private. You can generate a new key from inside your anvil app.", init_session: "If you pass a function to the init_session keyword parameter, it will be called after the uplink connection is established, but before any other interaction", quiet: "Set quiet to True to surpress connection output. Errors will still be displayed."}, anvil$helpLink: "/docs/uplink"} ["connect"]

#!defFunction(anvil.server [uplink],#,)!2: {$doc: "Disconnect your uplink script from your anvil app. Your script is then free to call `anvil.server.connect()` with the same uplink key or a new uplink key."} ["disconnect"]

#!defFunction(anvil.server [uplink],#,)!2: {$doc: "A useful shortcut to keep your Python script running. This allows your app to `anvil.server.call` functions inside your Python script. You can use any other way to keep the process alive in place of this function.", anvil$helpLink: "/docs/uplink/setting_up#connecting"} ["wait_forever"]




import anvil.server

from ._constants import SERVER_PREFIX, NOT_FOUND

PREFIX = SERVER_PREFIX + "row."
_make_refs = None  # Circular import


class _Batcher:
    _name = ""
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        return cls._instance

    def __init__(self):
        self._active = False
        self._updates = []
        self._buffer = {}
        self._func = PREFIX + self._name

    @property
    def active(self):
        return self._active

    def push(self, cap, update=False):
        self._updates.append((cap, update))

    def reset(self):
        self._active = False
        self._updates.clear()
        self._buffer.clear()

    def __enter__(self):
        if self._active:
            raise RuntimeError("nested batching is not suppported")
        self._active = True

    def get_args(self, updates):
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, traceback):
        updates = self._updates
        try:
            if exc_value is None and updates:
                anvil.server.call(self._func, self.get_args(updates))
                for cap, update in updates:
                    cap.send_update(update)
        finally:
            self.reset()


class BatchUpdate(_Batcher):
    _name = "batch_update"

    def push(self, cap, update):
        self._updates.append((cap, update))
        self._buffer.setdefault(cap, {}).update(update)

    def get_updates(self, cap):
        return self._buffer.get(cap, {})

    def read(self, cap, key):
        return self.get_updates(cap).get(key, NOT_FOUND)

    def get_args(self, updates):
        global _make_refs
        if _make_refs is None:
            from ._refs import make_refs  # circular import

            _make_refs = make_refs

        return [(cap, _make_refs(update)) for cap, update in updates]


class BatchDelete(_Batcher):
    _name = "batch_delete"

    def get_args(self, updates):
        return [cap for cap, _ in updates]


batch_update = BatchUpdate()
batch_delete = BatchDelete()



import anvil.tz
from anvil.server import Capability, unwrap_capability

from ._constants import CAP_KEY, NOT_FOUND, UNCACHED


def validate_cap(cap, view_key, table_id, row_id=NOT_FOUND):
    import json  # we don't need this on the client

    _, _, view_dict, narrowed, _ = unwrap_capability(cap, ["_", "t", Capability.ANY, Capability.ANY, Capability.ANY])
    assert view_dict == json.loads(view_key)
    assert table_id == str(view_dict["id"])
    if row_id is not NOT_FOUND:
        assert row_id == str(narrowed["r"])


def clean_local_datetime(d):
    if d.tzinfo is not None:
        offset = d.utcoffset().total_seconds()
    else:
        offset = anvil.tz.tzlocal().utcoffset(d).total_seconds()
    return d.replace(tzinfo=anvil.tz.tzoffset(seconds=offset))


# Serialization helpers
def check_serialized(self, local_data):
    self_id = id(self)
    serialized = local_data.get(self_id, False)
    local_data[self_id] = True
    return serialized


def init_view_data(view_key, g_table_data):
    return g_table_data.setdefault(view_key, {})


def init_spec_rows(g_view_data, table_spec, cache_spec=None):
    g_table_spec = g_view_data.get("spec")
    if g_table_spec is not None:
        pass
    elif table_spec is None or cache_spec is None:
        g_table_spec = g_view_data["spec"] = table_spec
    else:
        g_table_spec = g_view_data["spec"] = {"cols": table_spec["cols"], "cache": cache_spec}
    g_table_rows = g_view_data.setdefault("rows", {})
    return g_table_spec, g_table_rows


def merge_row_data(row_id, row_data, g_table_rows, g_table_spec, row_cache_spec):
    # we've already cleaned the row_data
    #  - it will only be a compact list if the caches match
    #  - and g_row_data is either None or also a compact list
    # otherwise row_data will be a dict
    g_row_data = g_table_rows.get(row_id)

    # FAST - common case - nothing in row_data
    if g_row_data is None:
        g_table_rows[row_id] = row_data
        return

    g_row_type = type(g_row_data)
    row_type = type(row_data)

    # handle all UNCACHED - i.e. the partially cached writer wins
    if g_row_type is list and len(g_row_data) == 1:
        # the row serialized before us has an all 0 cache_spec and is compact
        # we are either a dict or a list of the same length
        g_table_rows[row_id] = row_data
        return
    if not any(row_cache_spec):
        # the row to merge has an all 0 cache_spec
        return

    # SLOW PATH - uncommon cases
    # Another reference to this row (not the exact same row) was already serialized before us
    if row_type is list:
        # g_row_data must also be a compact list if row_data is a list
        # they must have the same length at this stage since we know the cache specs match
        if g_row_type is list:
            # fail safe sanity check
            merge_compact(row_data, g_row_data)
        
    elif g_row_type is dict:
        # then the previously serialized reference to this row
        # didn't match the g_cache_spec
        # so just take the itersect of the dictionaries
        g_table_rows[row_id] = merge_dicts(row_data, g_row_data)
        return
    else:
        # finally the g_row_type is a compact list and we are a dict - make it a dict
        g_cache_spec = g_table_spec["cache"]
        merge_dict_with_compact(row_data, g_row_data, row_cache_spec, g_cache_spec)
        g_table_rows[row_id] = row_data


def merge_compact(row_data, g_row_data):
    # any conflicts just replace with UNCACHED sentinel
    # use len - 1 so we skip the Capability
    for i in range(len(row_data) - 1):
        gbl, loc = g_row_data[i], row_data[i]
        if gbl != loc:
            g_row_data[i] = UNCACHED


def merge_dicts(row_data, g_row_data):
    # walk the smallest
    merged = {}
    a, b = (row_data, g_row_data) if len(row_data) < len(g_row_data) else (g_row_data, row_data)
    cap = a.pop(CAP_KEY)
    for key, a_val in a.items():
        b_val = b.get(key, NOT_FOUND)
        if a_val == b_val:
            merged[key] = a_val
    merged[CAP_KEY] = a[CAP_KEY] = cap
    return a


def merge_dict_with_compact(row_data, g_row_data, row_cache_spec, g_cache_spec):
    iter_g_row_data = iter(g_row_data)
    for i, (is_cached, g_is_cached) in enumerate(zip(row_cache_spec, g_cache_spec)):
        i = str(i)
        if not g_is_cached:
            # we could use the incoming caller wins here
            if is_cached:
                row_data.pop(i, None)
            continue

        g_val = next(iter_g_row_data)
        if not is_cached:
            continue

        if i in row_data and row_data[i] != g_val:
            row_data.pop(i)

    return row_data



# Helpers for implementing anvil.server.
# Used in uplink, downlink and pypy-sandbox.
import importlib

import anvil
import traceback
import numbers
import sys
import re
import json
import math
import anvil.tz
import functools

_do_call = None

string_type = str if sys.version_info >= (3,) else basestring
long_type = int if sys.version_info >= (3,) else long

POS_INFINITY = float("inf")
NEG_INFINITY = float("-inf")

_value_types = {}
_serialization_helpers = {} # {module_name: helper_fn}

class LiveObjectProxy(anvil.LiveObject):

    def __init__(self,spec,known_methods=None):
        for k in ["itemCache", "iterItems"]:
            if spec.get(k, {}) is None:
                del spec[k]

        if known_methods is not None:
            if spec.get("methods") is None:
                spec["methods"] = known_methods[spec["backend"]]
            else:
                known_methods[spec["backend"]] = spec["methods"]
        anvil.LiveObject.__init__(self, spec)

    def __getattr__(self, item):
        if item in self._spec["methods"]:
            def item_fn(*args, **kwargs):
                return _do_call(args, kwargs, fn_name=item, live_object=self)

            return item_fn
        else:
            raise AttributeError(item)

    def __getitem__(self, item):

        if "__anvil_iter_page__" in self._spec["methods"]:
            if isinstance(item, (int,long_type)):
                if item < 0:
                    raise IndexError("list index cannot be negative")

                iter = LiveObjectProxy.Iter(self, item)
                try:
                    return iter.next()
                except StopIteration:
                    raise IndexError("list index out of range")
            elif isinstance(item, slice):
                if (item.start and item.start < 0) or (item.stop and item.stop < 0) or (item.step and item.step < 0):
                    raise Exception("list slice indices and step cannot be negative")
                return LiveObjectProxy.Iter(self, item.start, item.stop, item.step)


        if item in self._spec.get("itemCache", {}):
            return self._spec["itemCache"][item]

        getitem = self.__getattr__("__getitem__")

        try:
            return getitem(item)
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)

    def __setitem__(self, key, value):
        if key in self._spec.get("itemCache", {}):
            del self._spec["itemCache"][key]

        setitem = self.__getattr__("__setitem__")
        try:
            r = setitem(key, value)
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)

        if "itemCache" in self._spec and (isinstance(value, string_type) or isinstance(value, numbers.Number) or isinstance(value, bool) or value is None):
            self._spec["itemCache"][key] = value

        return r

    class Iter:
        def __init__(self, live_object, start=None, stop=None, step=None):
            self._live_object = live_object

            i = live_object._spec.get("iterItems", {})
            self._idx = start if start is not None else 0
            self._items = i.get("items", None)
            self._next_page = i.get("nextPage", None)
            self._stop = stop
            self._step = step if step is not None else 1

        def _fetch_state(self):
            r = _do_call([self._next_page], {}, fn_name="__anvil_iter_page__", live_object=self._live_object)
            self._items = r["items"]
            self._next_page = r.get("nextPage", None)

        def __iter__(self):
            return self

        def next(self):
            if self._items is None:
                try:
                    self._fetch_state()
                except AnvilWrappedError as e:
                    raise _deserialise_exception(e.error_obj)

            if self._idx < len(self._items) and (self._stop is None or self._idx < self._stop):
                r = self._items[self._idx]
                self._idx += self._step
                return r

            if self._next_page is None or (self._stop is not None and self._idx >= self._stop):
                raise StopIteration

            self._idx -= len(self._items) if self._items is not None else self._idx
            if self._stop is not None:
                self._stop -= len(self._items) if self._items is not None else 0
            self._items = None
            return self.next()

        def __next__(self):
            return self.next()

    def __iter__(self):
        if "__anvil_iter_page__" in self._spec["methods"]:
            return LiveObjectProxy.Iter(self)
        else:
            raise Exception("Not iterable: <LiveObject: %s>" % self._spec.get("backend", "INVALID"))

    def __bool__(self):
        if "__nonzero__" in self._spec["methods"]:
            return self.__getattr__("__nonzero__")()
        else:
            return True

    __nonzero__ = __bool__

    def __len__(self):
        if "__len__" in self._spec["methods"]:
            return int(self.__getattr__("__len__")())
        else:
            l = 0
            for _ in self.__iter__():
                l += 1
            return l


# Wildcard for unwrap_capability
class _CapAny(object):
    def __repr__(self):
        return "ANY"

def _check_valid_scope(scope, name="scope"):
    if type(scope) is not list:
            raise TypeError("The {} of a Capabilty must be a list".format(name))
    try:
        return json.loads(json.dumps(scope))
    except TypeError as e:
        raise TypeError("The {} provided is not valid JSON data. {}".format(name, e))


class Capability(object):
    def __init__(self, scope, mac=None, narrow=None):
        scope = _check_valid_scope(scope)
        self._scope = scope
        self._mac = mac
        if mac is not None:
            pass
        elif not len(scope):
            raise ValueError("Cannot construct a capability with an empty scope")
        elif scope[0] == "_":
            raise ValueError("To construct a Capability from scratch, its scope cannot start with ['_']")

        self._narrow = narrow or []
        self._do_apply_update = None
        self._do_get_update = None
        self._queued_update = {}
        self._hash = None

    @property
    def scope(self):
        return self._scope + self._narrow

    def narrow(self, narrowing_suffix):
        narrowing_suffix = _check_valid_scope(narrowing_suffix, "narrow argument")
        return Capability(self._scope, self._mac, self._narrow + narrowing_suffix)

    def __repr__(self):
        return "<anvil.server.Capability:{}>".format(self.scope)

    def __eq__(self, other):
        if type(other) is not Capability:
            return NotImplemented
        return self.scope == other.scope

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(json.dumps(self.scope))
        return self._hash

    def set_update_handler(self, apply_update, get_update=None):
        self._do_apply_update = apply_update
        self._do_get_update = get_update

    # RPC machinery:
    # An update has arrived from a server call. Default behaviour: Merge if it's a dict, overwrite if it isn't
    def _apply_update(self, update):
        if self._do_apply_update is not None:
            self._do_apply_update(update)

        if self._do_get_update is None:
            # Default update propagation rules
            if isinstance(update, dict):
                self._queued_update.update(**update)
            else:
                self._queued_update = update

    send_update = _apply_update

    # RPC machinery:
    # We're about to return from a server call which passed this capability in; do we have an update for them?
    # None -> nothing to send
    def _get_update(self):
        if self._do_get_update is not None:
            return self._do_get_update()
        else:
            return None if self._queued_update == {} else self._queued_update

    # Sentinel value for unwrap_capability
    ANY = _CapAny()

#!defFunction(anvil.server,list,capability,scope_pattern)!2: "Checks that its first argument is a valid Capability, and that its scope matches the supplied pattern.\n\nTo match, the scope must:\n - Be at least as broad as the pattern (ie the same length or shorter)\n- Contain the same values in the same position as the pattern - unless that position in the pattern is Capability.ANY, which matches any value\n\nReturns a list of matched scope elements, of the same length as the pattern. (If the scope was broader than required, missing elements are set to None.)" ["unwrap_capability"]
def unwrap_capability(cap, scope_pattern):
    if type(cap) is not Capability:
        raise TypeError("Not a valid Capability: found {}".format(type(cap).__name__))
    if type(scope_pattern) is not list:
        raise TypeError("scope_pattern should be a list, not {}".format(type(scope_pattern).__name__))

    scope = cap.scope
    ret = [None] * len(scope_pattern)

    if len(scope) > len(scope_pattern):
        raise ValueError("Capability is too narrow: required %s; got %s" % (scope_pattern, scope))

    for i in range(len(scope)):
        if scope_pattern[i] is Capability.ANY or scope[i] == scope_pattern[i]:
            ret[i] = scope[i]
        else:
            raise ValueError("Incorrect Capability: required %s; got %s" % (scope_pattern, cap.scope))

    return ret

# DEPRECATED: replaced by unwrap_capability - included here for Backwards compatibility
Capability.require = staticmethod(unwrap_capability)


class SerializationInfo(object):
    def __init__(self, fromdata=None, remote_is_trusted=False):
        self._txdata = {}
        self._localdata = {}
        self._defaultkey = None
        self._trusted = remote_is_trusted
        self._enable_txdata = True
        if fromdata is None:
            pass
        elif type(fromdata) is dict:
            self._txdata[":GLOBAL"] = fromdata
        else:
            it = iter(fromdata)
            for k, v in zip(it, it):
                self._txdata[k] = v

    def __bool__(self):
        return bool(self._enable_txdata)

    __nonzero__ = __bool__

    def _to_json(self):
        if len(self._txdata) == 1 and ":GLOBAL" in self._txdata:
            return self._txdata[":GLOBAL"]
        else:
            r = []
            for k, v in self._txdata.items():
                r.append(k)
                r.append(v)
            return r

    def _resolve_key(self, key):
        if key is None:
            return self._defaultkey
        elif isinstance(key, type):
            try:
                return key.SERIALIZATION_INFO[0]
            except AttributeError:
                return key.__module__ + "." + key.__name__
        else:
            return ":" + str(key)

    def _set_default_key(self, key):
        self._defaultkey = key

    def _set_txdata_available(self, enable):
        self._enable_txdata = enable
    
    def _set_data_factory(self, _data, resolved_key, factory):
        data = _data.get(resolved_key)
        if data is None:
            data = _data[resolved_key] = factory()
        return data

    def shared_data(self, key=None, transmitted_data_factory=dict, local_data_factory=dict):
        key = self._resolve_key(key)
        localdata = self._set_data_factory(self._localdata, key, local_data_factory)
        if not self._enable_txdata:
            return None, localdata
        txdata = self._set_data_factory(self._txdata, key, transmitted_data_factory)
        return txdata, localdata

    @property
    def remote_is_trusted(self):
        return self._trusted

    @property
    def local_is_trusted(self):
        return anvil.is_server_side()

    def __repr__(self):
        # TODO adjust for public api
        return "SerializationInfo<" + repr(self._txdata) + ", " + repr(self._localdata) + ">"


# Backwards compatibility: If you use SerializationInfo like a dict, it man works like  if you use it like a dict
def _wrap_global_dict(attr_name):
    def f(self, *args, **kws):
        transmitted_data = self.shared_data("GLOBAL")[0]
        if transmitted_data is None:
            # using the old API so better to throw here
            raise RuntimeError("This object is part of shared_data; you cannot access shared_data from its __serialize__ method.")
        return getattr(transmitted_data, attr_name)(*args, **kws)
    return f


for method in ["__getitem__", "__setitem__", "__delitem__", "__iter__", "__len__", "__contains__", "keys", "items",
               "values", "get", "pop", "popitem", "clear", "update", "setdefault"]:
    setattr(SerializationInfo, method, _wrap_global_dict(method))


# DEPRECATED: there is no longer any reason to inherit from this
class Serializable(object):
    SERIALIZATION_INFO = None

    def __serialize__(self, info):
        return self.__dict__

    # You only need one of __deserialize__ (called instead of __init__)
    # or __new_deserialized__ (called instead of __new__+__init__).
    # (Of the two, you almost always want __deserialize__.)
    def __deserialize__(self, from_data, info):
        self.__dict__.update(from_data)

    @classmethod
    def __new_deserialized__(cls, from_data, info):
        obj = cls.__new__(cls)
        obj.__dict__.update(from_data)
        return obj


# Add-in for maintaining object identity across serialization
class SerializeWithIdentity(object):
    @classmethod
    def __new_deserialized__(cls, from_data, global_data):
        clsname, _ = cls.SERIALIZATION_INFO
        cache = global_data.get(clsname)
        if cache is None:
            cache = global_data[clsname] = {}
        my_id = from_data[0]
        # We use global_data to cache instances we're constructing
        # (non-JSONable keys certainly won't conflict with other global_data users)
        obj = cache.get(my_id)
        if obj is None:
            obj = cache[my_id] = cls.__new__(cls)

        if len(from_data) > 1:
            obj.__deserialize__(from_data[1], global_data)

        return obj

    def __serialize__(self, global_data):
        my_id, gd = self._serialization_key if hasattr(self, "_serialization_key") else (None, None)
        if gd is global_data:
            return [my_id]

        clsname, _ = self.SERIALIZATION_INFO
        my_id = global_data.get(clsname+"_max", 0)
        global_data[clsname+"_max"] = my_id + 1

        self._serialization_key = (my_id, global_data)

        data = self.__serialize_once__(global_data)
        if isinstance(data, dict):
            data = dict(data)
            data.pop("_serialization_key", None)

        return [my_id, data]


def portable_class(cls, name=None):
    def register(cls, name):
        if not hasattr(cls, "__new__"):
            raise TypeError("Portable classes must be new-style classes (inherit from object). %s is not a new-style class." % repr(cls))
        if name is None:
            name = cls.__module__ + "." + cls.__name__
        elif not isinstance(name, str):
            raise TypeError("The second argument to portable_class must be a str")
        _value_types[name] = cls
        cls.SERIALIZATION_INFO = (name, cls)
        return cls

    if name is None and isinstance(cls, str):
        name = cls
        return lambda cls: register(cls, name)
    else:
        return register(cls, name)


# Old name, for apps written before portable classes were released
serializable_type = portable_class


class LazyMedia(anvil.Media):
    def __init__(self, spec):
        if isinstance(spec,LazyMedia):
            spec = spec._spec
        self._spec = spec
        self._details = None
        self._fetched = None

    def _fetch(self):
        if self._details is None:
            import anvil.server
            self._fetched = anvil.server.call("anvil.private.fetch_lazy_media", self._spec)
        return self._fetched

    def _get(self, key, attr=None):
        if attr is None:
            attr = key
        return self._spec[key] if key in self._spec else getattr(self._fetch(), attr)

    def get_name(self):
        try:
            return self._get("name")
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)

    def get_url(self, download=True):
        return anvil.server.call("anvil.private.get_lazy_media_url", self, download)

    def get_content_type(self):
        try:
            return self._get("mime-type", "content_type")
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)

    def get_length(self):
        try:
            return self._get("length")
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)

    def get_bytes(self):
        try:
            return self._fetch().get_bytes()
        except AnvilWrappedError as e:
            raise _deserialise_exception(e.error_obj)


class AnvilWrappedError(Exception):
    registered_type_name = None

    def __init__(self, message=""):
        if isinstance(message, dict):
            self.manually_created = False
            self.error_obj = message
        else:
            self.manually_created = True
            self.error_obj = {"message": str(message), "trace": []}
            t = type(self).registered_type_name
            if t is not None:
                self.error_obj["type"] = t
        self.message = self.error_obj.get("message", "")
        Exception.__init__(self, self.message)

    def __repr__(self):
        r = Exception.__repr__(self)
        if type(self) is not AnvilWrappedError:
            return r
        eo_type = self.error_obj.get("type")
        if eo_type is None:
            return r

        return "AnvilWrappedError" + "(" + eo_type + r[len("AnvilWrappedError"):] + ")"


def augment_exception(exc_class):
    def f(error_obj):
        e = exc_class(error_obj['message'])
        e._anvil_error_obj = error_obj
        return e
    return f

_named_exceptions = {
    "KeyError": augment_exception(KeyError)
}


def _register_exception_type(name, cls):
    _named_exceptions[name] = cls
    cls.registered_type_name = name


def _deserialise_exception(error_obj):
    return _named_exceptions.get(error_obj.get("type"), AnvilWrappedError)(error_obj)


def get_liveobject_cache_filter_spec(input_data):
    """Returns a specification of a filter for LiveObject cache updates, which passes only LiveObjects present in input_data"""
    spec = {}

    def f(data):
        if isinstance(data, list):
            for i in data: f(i)
        elif isinstance(data, dict):
            for k,v in data.items(): f(v)
        elif isinstance(data, LiveObjectProxy):
            b = spec.get(data._spec["backend"])
            if b is None:
                b = spec[data._spec["backend"]] = set()
            b.add(data._spec["id"])

    f(input_data)
    return spec


def combine_cache_updates(updates, new_updates, filter_spec):
    if updates is None:
        updates = {}

    for k in new_updates:
        permitted_ids = filter_spec.get(k)
        if permitted_ids is None:
            continue
        caches = updates.get(k)
        if caches is None:
            caches = updates[k] = dict()
        for id in new_updates[k]:
            if id not in permitted_ids:
                continue
            caches[id] = new_updates[k][id]


def apply_cache_updates(cache_updates, objects_to_walk):
    """Recursively walk an object tree, applying cache updates wherever necessary"""

    def f(data):
        if isinstance(data, list):
            for i in data: f(i)
        elif isinstance(data, dict):
            for k,v in data.items(): f(v)
        elif isinstance(data, LiveObjectProxy):
            update = cache_updates.get(data._spec["backend"], {})
            if data._spec["id"] in update:
                data._spec["itemCache"] = update[data._spec["id"]]

    f(objects_to_walk)


class MaybeWrappedError(Exception):
    def __init__(self, message=""):
        if isinstance(message, dict):
            Exception.__init__(self, message["message"])
        else:
            Exception.__init__(self, message)


class SerializationError(MaybeWrappedError):
    pass


class InternalError(MaybeWrappedError):
    pass


class InvalidResponseError(MaybeWrappedError):
    pass


class RuntimeUnavailableError(MaybeWrappedError):
    pass


class UplinkDisconnectedError(MaybeWrappedError):
    pass


class ExecutionTerminatedError(MaybeWrappedError):
    pass


class TimeoutError(MaybeWrappedError):
    pass


class QuotaExceededError(MaybeWrappedError):
    pass


class NoServerFunctionError(AnvilWrappedError):
    pass


class CookieError(AnvilWrappedError):
    pass


class _FailError(MaybeWrappedError):
    pass


class BackgroundTaskError(MaybeWrappedError):
    pass


class BackgroundTaskNotFound(MaybeWrappedError):
    pass


class BackgroundTaskKilled(MaybeWrappedError):
    pass


class PermissionDenied(MaybeWrappedError):
    pass


class ServiceNotAdded(MaybeWrappedError):
    pass



_register_exception_type("anvil.server.SerializationError", SerializationError)
_register_exception_type("anvil.server.InternalError", InternalError)
_register_exception_type("anvil.server.InvalidResponseError", InvalidResponseError)
_register_exception_type("anvil.server.RuntimeUnavailableError", RuntimeUnavailableError)
_register_exception_type("anvil.server.UplinkDisconnectedError", UplinkDisconnectedError)
_register_exception_type("anvil.server.ExecutionTerminatedError", ExecutionTerminatedError)
_register_exception_type("anvil.server.TimeoutError", TimeoutError)
_register_exception_type("anvil.server.QuotaExceededError", QuotaExceededError)
_register_exception_type("anvil.server.NoServerFunctionError", NoServerFunctionError)
_register_exception_type("anvil.server.CookieError", CookieError)
_register_exception_type("anvil.server._FailError", _FailError)
_register_exception_type("anvil.server.BackgroundTaskError", BackgroundTaskError)
_register_exception_type("anvil.server.BackgroundTaskNotFound", BackgroundTaskNotFound)
_register_exception_type("anvil.server.PermissionDenied", PermissionDenied)
_register_exception_type("anvil.server.ServiceNotAdded", ServiceNotAdded)



def _report_exception(request_id=None):
    exc_type, exc_value, exc_traceback = sys.exc_info()
    tb = traceback.extract_tb(exc_traceback)

    trace = [(filename.replace("\\","/"), lineno) for (filename, lineno, _, _) in tb]
    trace.reverse()

    # Last element of trace is where we called into user code. Remove it.
    trace.pop()

    if isinstance(exc_value, AnvilWrappedError):
        if not exc_value.manually_created:
            # First element of trace is where we re-raised the exception. Remove it.
            trace = trace[1:]
        exc_value.error_obj["trace"] = exc_value.error_obj.get("trace", []) + trace
        r = {
            "error": exc_value.error_obj,
            "id": request_id
        }
        return r
    elif isinstance(exc_value, SyntaxError):
        # Remove whole internal trace and replace it with line where error occurred.
        trace=[(exc_value.filename, exc_value.lineno)]
        return {
            "error": {
                "type": "SyntaxError",
                "trace": trace,
                "message": str(exc_value)
            },
            "id": request_id
        }
    elif isinstance(exc_value, MaybeWrappedError):
        return {
            "error": {
                "type": "anvil.server.%s" % exc_type.__name__,
                "trace": trace,
                "message": str(exc_value),
            },
            "id": request_id
        }
    elif hasattr(exc_value, "_anvil_error_obj"):
        error_obj = dict(exc_value._anvil_error_obj)
        error_obj['trace'] = exc_value._anvil_error_obj.get("trace", []) + trace[1:]

        return {
            "error": error_obj,
            "id": request_id
        }
    else:
        return {
            "error": {
                "type": str(exc_type.__name__),
                "trace": trace,
                "message": str(exc_value),
            },
            "id": request_id
        }

def reconstruct_val(v, known_liveobject_methods, reconstruct_data_media=None):

    for t in v["type"]:
        if t == "DataMedia":
            if reconstruct_data_media is None:
                raise Exception("No data media deserialiser provided. Cannot reconstruct.")
            return reconstruct_data_media(v)
        elif t == "LazyMedia":
            return LazyMedia(v)
        elif t == "LiveObject":
            return reconstruct_live_object(v, known_liveobject_methods)
        elif t == "Capability":
            return Capability(v["scope"], v["mac"])
        elif t == "Date":
            return parsedate(v["value"]) if v["value"] else None
        elif t == "DateTime":
            return parsedatetime(v["value"]) if v["value"] else None
        elif t == "Long":
            return long_type(v["value"])
        elif t == "Float":
            return float(v["value"])
        elif t == "Primitive":
            return v["value"]
        elif t == "ValueType":
            return v["typeName"]
        elif t == "ClassType":
            return v["typeName"]

    raise Exception("Server module cannot accept an object of type '%s'" % v["type"][0])



def reconstruct_live_object(d, known_methods):

    # Need to fill out known_methods before we can reconstruct members:
    if d.get('methods') is not None:
        known_methods[d["backend"]] = d["methods"]

    reconstructed_item_cache = {}
    for k,v in d.get("itemCache", {}).items():
        reconstructed_item_cache[k] = reconstruct_val(v, known_methods)
    d["itemCache"] = reconstructed_item_cache

    if d.get("iterItems"):
        reconstructed_iteritems = []
        for i in d["iterItems"]["items"]:
            reconstructed_iteritems.append(reconstruct_val(i, known_methods))
        d["iterItems"]["items"] = reconstructed_iteritems

    return LiveObjectProxy(d, known_methods)


def serialise_val(v, known_liveobject_methods):
    import datetime
    if isinstance(v, long_type) and not isinstance(v, bool):
        return {
            "type": ["Long"],
            "value": str(v)
        }
    elif isinstance(v, float) and (v == POS_INFINITY or v == NEG_INFINITY or math.isnan(v)):
        return {
            "type": ["Float"],
            "value": "Infinity" if v == POS_INFINITY else \
                        "-Infinity" if v == NEG_INFINITY else \
                        "NaN"
        }
    elif isinstance(v, (numbers.Number, bool, string_type)) or v is None:
        return {
            "type": ["Primitive"],
            "value": v
        }
    elif isinstance(v, anvil.LiveObject):
        return serialise_live_object(v, known_liveobject_methods)
    elif isinstance(v, datetime.datetime):
        s = "%04d-%02d-%02d %02d:%02d:%02d.%06d" % (v.year, v.month, v.day, v.hour, v.minute, v.second, v.microsecond)

        if v.tzinfo is not None:
            offset = v.tzinfo.utcoffset(v).total_seconds()
        else:
            offset = anvil.tz.tzlocal().utcoffset(v).total_seconds()

        sign = "+" if offset >= 0 else "-"
        z = "%s%02d%02d" % (sign, abs(int(offset/3600)), int((offset % 3600)/60))

        return {
            "type": ["DateTime"],
            "value": s + z
        }
    elif isinstance(v, datetime.date):
        s = "%04d-%02d-%02d" % (v.year, v.month, v.day)
        return {
            "type": ["Date"],
            "value": s
        }
    elif isinstance(v, LazyMedia):
        return dict(v._spec)
    else:
        for [n, cls] in _value_types.items():
            if isinstance(v, cls):
                type_name = n
                break
        else:
            raise Exception("Cannot serialise object of type %s" % type(v).__name__)

        return {
            "type": ["ValueType"],
            "typeName": type_name
        }


def serialise_live_object(obj, known_methods):
    obj = obj._spec.copy()
    obj["type"] = ["LiveObject"]

    km = known_methods.get(obj["backend"])
    if obj["methods"] == km:
        del obj["methods"]
    else:
        known_methods[obj["backend"]] = obj["methods"]

    serialised_item_cache = {}
    for k,v in sorted(obj.get("itemCache", {}).items(), key=lambda x: x[0]):
        # Dictionaries aren't ordered, so we can't allow this element
        # to write into known_methods. We give it a copy instead.
        serialised_item_cache[k] = serialise_val(v, known_methods.copy())
    obj["itemCache"] = serialised_item_cache

    if obj.get("iterItems"):
        serialised_iteritems = []
        for i in obj["iterItems"]["items"]:
            serialised_iteritems.append(serialise_val(i, known_methods))
        obj["iterItems"]["items"] = serialised_iteritems

    return obj


def _repr_path(p):
    return "".join(("[%s]" % repr(k) for k in p))

def _module_prefixes(module):
    module_parts = module.split(".")
    return [".".join(module_parts[:i]) for i in range(1, len(module_parts)+1)]

_called_serialization_helpers = set()

def _check_and_call_serialization_helper(cls_fullname):
    "checks if a class has a registered helper, calls the helper if it exists (once)"
    if cls_fullname in _called_serialization_helpers:
        return False

    for prefix in _module_prefixes(cls_fullname):
        if prefix in _serialization_helpers:
            _serialization_helpers[prefix](cls_fullname)
            _called_serialization_helpers.add(cls_fullname)
            return True

    return False


def fill_out_media(json, handle_media_fn, collect_capabilities=None, remote_is_trusted=False):
    obj_descr = []
    path = []
    known_liveobject_methods = {}
    serialization_info = SerializationInfo(remote_is_trusted=remote_is_trusted)
    import datetime

    def do_fom(_json):

        t_json = type(_json)

        if hasattr(_json, "SERIALIZATION_INFO"):
            type_name, tp = _json.SERIALIZATION_INFO
            valid_type_name = type_name in _value_types

            if valid_type_name and tp is _json:
                _json = None
                obj_descr.append({
                    "type": ["ClassType"],
                    "path": list(path),
                    "typeName": type_name
                })

            elif not valid_type_name or t_json is not tp:
                raise SerializationError("Cannot serialize %s (must be registered with @anvil.server.portable_class) at msg%s" % (t_json, _repr_path(path)))
            else:
                serialization_info._set_default_key(type_name)

                try:
                    serialize = _json.__serialize__
                except AttributeError:
                    def serialize(_):
                        return _json.__dict__

                content = serialize(serialization_info)

                _json = do_fom(content)

                # Append this afterwards, so we deserialise our content first
                obj_descr.append({
                    "type": ["ValueType"],
                    "path": list(path),
                    "typeName": type_name
                })

        elif isinstance(_json, dict):
            _json = dict(_json)
            for k,v in sorted(_json.items(),key=lambda x: x[0]):
                if not isinstance(k, string_type):
                    raise SerializationError("Cannot serialize dictionaries with keys that aren't strings at msg%s" % _repr_path(path + [k]))
                path.append(k)
                _json[k] = do_fom(v)
                path.pop()
        elif isinstance(_json, list) or isinstance(_json, tuple):
            _json = list(_json)
            for i in range(len(_json)):
                path.append(i)
                _json[i] = do_fom(_json[i])
                path.pop()
        elif isinstance(_json, LazyMedia):
            d = dict(_json._spec)
            d["path"] = list(path)
            obj_descr.append(d)
            _json = None
        elif isinstance(_json, anvil.Media):
            extra = handle_media_fn(_json)
            d = {"type": ["DataMedia"], "path": list(path), "mime-type": _json.content_type, "name": _json.name}
            if extra is not None:
                d.update(extra)
            obj_descr.append(d)
            _json = None
        elif isinstance(_json, Capability):
            if collect_capabilities is not None:
                collect_capabilities.append(_json)
            d = {
                "type": ["Capability"],
                "path": list(path),
                "scope": _json._scope,
                "mac": _json._mac
            }
            if _json._narrow:
                d["narrow"] = _json._narrow
            obj_descr.append(d)
            _json = None
        elif isinstance(_json, anvil.LiveObject):
            #print "Serialising LiveObject: " + repr(_json._spec) + " at " + repr(path)
            serialised_liveobject = serialise_live_object(_json, known_liveobject_methods)
            serialised_liveobject["path"] = list(path)
            obj_descr.append(serialised_liveobject)
            _json = None
        elif isinstance(_json, (datetime.date, datetime.datetime)) or \
                (isinstance(_json, long_type) and (_json > 2147483647 or _json < -2147483647)) or \
                (isinstance(_json, float) and (_json == POS_INFINITY or _json == NEG_INFINITY or math.isnan(_json))):
            serialised_val = serialise_val(_json, known_liveobject_methods)
            serialised_val["path"] = list(path)
            obj_descr.append(serialised_val)
            _json = None
        elif _check_and_call_serialization_helper(t_json.__module__ + "." + t_json.__name__):
            _json = do_fom(_json)
        elif 'numpy' in sys.modules and hasattr(sys.modules['numpy'], 'generic') and isinstance(_json, sys.modules['numpy'].generic):

            _json = _json.item() # convert

        elif not (isinstance(_json, string_type) or _json is None or _json is True or _json is False or isinstance(_json, (int, long_type)) or isinstance(_json, float)):

            # Rescue: Convert numpy types to nearest equivalent
            if 'numpy' in sys.modules:
                import numpy
                if isinstance(_json, numpy.ndarray):
                    _json = do_fom(_json.tolist())
                else:
                    raise SerializationError("Cannot serialize %s object at msg%s" % (t_json, _repr_path(path)))
            else:
                raise SerializationError("Cannot serialize %s object at msg%s" % (t_json, _repr_path(path)))

        return _json

    json = do_fom(json)

    vt_global = serialization_info._to_json()
    if len(vt_global) != 0:
        path.append("vt_global")
        serialization_info._set_txdata_available(False)
        od = obj_descr
        obj_descr = []
        json["vt_global"] = do_fom(vt_global)
        obj_descr += od
        path.pop()

    json["objects"] = obj_descr

    return json


def fill_out_cap_updates(resp, caps_passed_in):
    """We're about to return from a call; ask all the capabilities that got passed in whether they want to send
       any updates to our caller."""

    for cap in caps_passed_in:
        update = cap._get_update()
        if update is not None:
            cu = resp.get('capUpdates')
            if cu is None:
                cu = resp['capUpdates'] = {}
            cu[json.dumps(cap.scope)] = update


def apply_cap_updates(resp, caps_passed_out):
    """We have just made a server call that has returned. Apply any necessary updates to the capabilities we
       passed into this call"""

    updates = resp.get('capUpdates', {})

    # Normalise updates to how _this_ `json` impl does things (ugh)
    updates = {json.dumps(json.loads(k)): v for k,v in updates.items()}

    for cap in caps_passed_out:
        scope_json = json.dumps(cap.scope)
        update = updates.get(scope_json)
        if update is not None:
            cap._apply_update(update)


def simple_strpdate(s):
    import datetime
    return datetime.date(int(s[0:4]), int(s[5:7]), int(s[8:10]))


def simple_strpdatetime(s):
    import datetime
    return datetime.datetime(int(s[0:4]), int(s[5:7]), int(s[8:10]), int(s[11:13]), int(s[14:16]), int(s[17:19]), int(s[20:26])) # datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S.%f")

def parsedate(s):
    return simple_strpdate(s)


def parsedatetime(s):

    has_tz = len(s) > 5 and \
             (s[-5] == "-" or s[-5] == "+") and \
             48 <= ord(s[-4]) <= 57 and \
             48 <= ord(s[-3]) <= 57 and \
             48 <= ord(s[-2]) <= 57 and \
             48 <= ord(s[-1]) <= 57

    if not has_tz:
        # Parse a naive datetime
        return simple_strpdatetime(s)

    # Timezone present. First parse without it
    d = simple_strpdatetime(s[:-5])

    # Now construct a tzoffset
    hours = int(s[-5:-2])
    minutes = int(s[-5] + s[-2:])
    total_minutes = hours*60+minutes

    return d.replace(tzinfo=anvil.tz.tzoffset(minutes=total_minutes))

def _retrieve_portable_class(type_name, d):
    value_type = _value_types.get(type_name)
    if value_type is not None:
        return value_type
    # Try importing the relevant module
    i = type_name.rfind('.')
    if i != -1:
        # TODO do we filter what we can specify as import? I don't *think* this is dangerous...
        if not _check_and_call_serialization_helper(type_name):
            module_name = type_name[:i]
            importlib.import_module(module_name)
        value_type = _value_types.get(type_name)

    if value_type is None:
        raise SerializationError("No such serializable type: %s at msg%s" % (type_name, _repr_path(d["path"])))

    return value_type


def _reconstruct_objects(json, reconstruct_data_media, hold_back_value_types=False, collect_capabilities=None, remote_is_trusted=False):
    known_liveobject_methods = {}
    serialization_info = SerializationInfo(json.get("vt_global"), remote_is_trusted=remote_is_trusted) if not hold_back_value_types else None

    if "objects" in json:
        held_back_objects = []
        for d in json["objects"]:
            if hold_back_value_types and ("ValueType" in d["type"] or "ClassType" in d["type"]):
                held_back_objects.append(d)
                continue

            reconstructed = reconstruct_val(d, known_liveobject_methods, reconstruct_data_media)
            if collect_capabilities is not None and type(reconstructed) is Capability:
                collect_capabilities.append(reconstructed)

            obj = json
            last_obj = None
            key = None
            for k in d["path"]:
                last_obj = obj
                key = k
                obj = obj[k]

            if last_obj is not None:

                if "ValueType" in d["type"]:
                    # Hack: The "reconstructed value" here is actually just the type name
                    type_name = reconstructed
                    value_type = _retrieve_portable_class(type_name, d)
                    try:
                        reconstruct = value_type.__new_deserialized__
                    except AttributeError:
                        def reconstruct(data, info):
                            obj = value_type.__new__(value_type)
                            try:
                                deserialize = obj.__deserialize__
                            except AttributeError:
                                def deserialize(data, info):
                                    obj.__dict__.update(data)
                            deserialize(data, info)
                            return obj

                    serialization_info._set_txdata_available(d["path"][0] != 'vt_global')
                    serialization_info._set_default_key(type_name)
                    last_obj[key] = reconstruct(last_obj[key], serialization_info)
                elif "ClassType" in d["type"]:
                    last_obj[key] = _retrieve_portable_class(reconstructed, d)
                else:
                    last_obj[key] = reconstructed

        if hold_back_value_types:
            json["objects"] = held_back_objects
        else:
            del json["objects"]

    return json

on_register = None # optional
registrations = {}

class HttpRequest(object):

    def __init__(self):
        self._prevent_access = True

    def __getattribute__(self, name):
        if object.__getattribute__(self, "_prevent_access"):
            raise Exception("anvil.server.request is only available in http_endpoint calls.")

        return object.__getattribute__(self, name)

    @property
    def body_json(self):
        if hasattr(self, "_body_json"):
            return self._body_json
        elif self.body is not None and self.headers.get("content-type", None) == "application/json":
            self._body_json = json.loads(self.body.get_bytes())
        else:
            self._body_json = None
        return self._body_json




api_request = HttpRequest()

def _lower_str(s):
    if not isinstance(s, str):
        return s
    return s.lower()

class HttpHeaders(object):

    def __init__(self, headers=None):
        headers = headers or {}
        assert isinstance(headers, dict), "headers should be a dict"
        headers = {_lower_str(h): v for (h, v) in headers.items()}
        self._headers = list(headers.items())

    def __setitem__(self, name, val):
        name = _lower_str(name)
        self._headers = [(h,v) for (h,v) in self._headers if h != name]

        self.add(name, val)

    def __delitem__(self, name):
        name = _lower_str(name)
        self._headers = [(h,v) for (h,v) in self._headers if h != name]

    def add(self, name, val):
        self._headers.append((name, val))

    def clear(self):
        self._headers = []

    def copy(self):
        new_self = HttpHeaders()
        new_self._headers = self._headers.copy()
        return new_self

    def __repr__(self):
        return "HttpHeaders(" + repr(self._headers) + ")"


class HttpResponse(object):
    def __init__(self, status=200, body="", headers=None):
        self.status = status
        self.body = body
        self.headers = headers

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value):
        if value is None or isinstance(value, dict):
            self._headers = HttpHeaders(value)
        elif isinstance(value, HttpHeaders):
            self._headers = value.copy()
        else:
            raise TypeError("headers should be set to a dictionary")


class CallContext(object):
    class ClientInfo(object):
        def __repr__(self):
            return "<ClientInfo:%s>" % repr(self.__dict__)

    class Location(object):
        def __init__(self, location):
            self.city = location.get('city')

            self.subdivision = location.get('subdivision', {}).get('name')

            l = location.get('location', {})
            self.latitude = l.get('lat')
            self.longitude = l.get('lng')

            c = location.get('country', {})
            self.country_code = c.get('code')
            self.country = c.get('name')

        def __repr__(self):
            return "<Location:%s>" % repr(self.__dict__)

    class StackFrame(object):
        def __init__(self, sf):
            self.type = sf.get("type")
            self.is_trusted = self.type in {'server_module', 'uplink', 'background_task'}

        def __repr__(self):
            return "<StackFrame:%s>" % repr(self.__dict__)

    _DEFAULT_TYPE = "server_module"

    def _setup(self, client, call_stack):
        self.type = CallContext._DEFAULT_TYPE
        if client is not None:
            self.client = CallContext.ClientInfo()
            self.client.type = client.get('type')
            self.client.ip = client.get('ip')
            self.background_task_id = client.get('background-task-id')
            l = client.get('location')
            self.client.location = CallContext.Location(l) if l else None
            self.remote_caller = CallContext.StackFrame(call_stack[0] if call_stack else client)
        else:
            self.client = None
            self.remote_caller = None
            self.background_task_id = None

    __init__ = _setup

    def __repr__(self):
        return "<CallContext:%s>" % repr(self.__dict__)


# can be used as a decorator too
# N.B. There is a full implementation of the 'require=' kwarg here, but we've chosen not to expose it yet. If we expose it, will need to pass through from callable, like require_user.
def register(fn, name=None, name_prefix=None, require_user=None):
    require=None

    if isinstance(fn, string_type):
        # Someone's using the old syntax. Our bad.
        (fn, name) = (name, fn)

    if name is None:
        name = fn.__name__

    original_name = name
    if name_prefix is not None:
        name = "%s:%s" % (name_prefix, name)

    # 'require_user' is either True (check that user is logged in), False (equivalent to None), or 
    # a function that takes the currently logged in user and returns whether to let them in.

    if require_user == True:
        def simple_require_user():
            import anvil.users
            if anvil.users.get_user() is None:
                raise anvil.users.AuthenticationFailed("You must be logged in to call this server function")
            return True
        require = simple_require_user
    elif require_user == False:
        require = None
    elif require_user is not None:
        # Must be a function.
        def complex_require_user():
            import anvil.users
            user = anvil.users.get_user()
            if user is None:
                raise anvil.users.AuthenticationFailed("You must be logged in to call server function '%s'" % original_name)
            elif not require_user(user):
                raise PermissionDenied("You do not have permission to call server function '%s'" % original_name)
            else:
                return True
        require = complex_require_user

    # 'require' is an optional function that returns something truthy if the user should be let in. Otherwise, it can raise 
    # an Exception (which will be passed on) or return something falsey (in which case a PermissionDenied exception will be raised.)

    if require is not None:
        def require_wrap(f):
            def with_req(*args, **kwargs):
                if require():
                    return f(*args, **kwargs)
                else:
                    raise PermissionDenied("You do not have permission to call server function '%s'" % original_name)
            return with_req
    else:
        def require_wrap(f):
            return fn

    registrations[name] = require_wrap(fn)

    if on_register is not None:
        on_register(name, False)

    def reregister(new_f):
        registrations[name] = require_wrap(new_f)
        new_f._anvil_reregister = reregister

    fn._anvil_reregister = reregister

    return fn

#!defFunction(anvil.server,%,[name], [require_user])!2: {anvil$args: {fn_or_name: "The name by which you want to call your function from the client.", require_user: "Allows you to verify whether a user is logged in. Can be a boolean or a function."}, anvil$helpLink: "server#calling-server-functions-from-client-code", $doc: "When applied to a function as a decorator, the function becomes available from the client side."} ["callable"]
def callable(fn_or_name=None, require_user=None):
    if fn_or_name is None or isinstance(fn_or_name, string_type):
        return lambda f: register(f, fn_or_name, require_user=require_user)
    else:
        return register(fn_or_name)


def background_task(fn_or_name=None):
    if fn_or_name is None or isinstance(fn_or_name, string_type):
        return lambda f: register(f, fn_or_name, name_prefix="task")
    else:
        return register(fn_or_name, name_prefix="task")


# A parameterised decorator
def callable_as(name):
    print("@callable_as is deprecated. Please use @callable directly.")
    return lambda f: register(f, name)



def http_endpoint(path, require_credentials=False, authenticate_users=False, authenticate_user=False,
                  methods=["GET","POST"], enable_cors=False, cross_site_session=False, _task_prefix="http"):
    def decorator(fn):
        path_parts = []
        def register_path_part(s):
            path_parts.append(s.group(1))
            return "([^/]*)"

        path_regex = re.sub(":([^/]*)", register_path_part, path)

        def wrapped_fn(method, path, query_params, form_params, origin, headers, remote_address, body, username, password, same_app_alternate_origin=None, **more_kwargs):

            api_request._prevent_access = False

            if cross_site_session:
                import anvil.server
                anvil.server._switch_session()

            api_request.user = None
            if authenticate_users or authenticate_user:
                import anvil.users
                try:
                    api_request.user = anvil.users.get_user() or \
                                        anvil.users.login_with_email(username, password)
                except Exception as e:
                    return {"status": 401,
                            "body": "Unauthorized",
                            "headers": {"WWW-Authenticate": "Basic realm=\"Anvil App API\""}}
            elif require_credentials:
                if username is None or password is None:
                    return {"status": 401,
                            "body": "Unauthorized",
                            "headers": {"WWW-Authenticate": "Basic realm=\"Anvil App API\""}}

            def add_cross_origin_to_header_dict(d):
                if enable_cors:
                    d["Access-Control-Allow-Origin"] = enable_cors if isinstance(enable_cors, str) else "*"
                elif same_app_alternate_origin:
                    d["Access-Control-Allow-Origin"] = same_app_alternate_origin
                if enable_cors or same_app_alternate_origin and \
                        "access-control-allow-headers" not in [h.lower() for h in d.keys()]:
                    allow_headers = ["content-type"]
                    for k in headers.keys():
                        if k.lower() == "access-control-request-headers":
                            hs = [h.strip().lower() for h in headers[k].split(",")]
                            allow_headers += [h for h in hs if h not in allow_headers]

                    d["Access-Control-Allow-Headers"] = ", ".join(allow_headers)

                return d

            if method not in methods:
                if method == "OPTIONS" and (enable_cors or same_app_alternate_origin):
                    return {"status": 200,
                            "body": "",
                            "headers": add_cross_origin_to_header_dict({})}
                else:
                    return {"status": 405,
                            "body": "Method not supported",
                            "headers": {"Allow": ", ".join(methods)}}

            api_request.path = path
            api_request.method = method
            api_request.query_params = query_params
            api_request.form_params = form_params
            api_request.origin = origin
            api_request.headers = headers
            api_request.remote_address = remote_address
            api_request.body = body
            api_request.username = username
            api_request.password = password


            # Path takes precedence over query params. Query params take precedence over form params.
            kwargs = dict(form_params)
            kwargs.update(query_params)
            match = re.match(path_regex, path)
            for i,m in enumerate(match.groups()):
                kwargs[path_parts[i]] = m

            response = fn(**kwargs)

            api_request._prevent_access = True

            if isinstance(response, HttpResponse):
                return {"status":  response.status,
                        "body":    response.body,
                        "headers": add_cross_origin_to_header_dict(response.headers)._headers}
            else:
                return {"status":  200,
                        "body":    response,
                        "headers": add_cross_origin_to_header_dict({})}

        register(wrapped_fn, path_regex, _task_prefix)

        return fn
    return decorator


wellknown_endpoint = functools.partial(http_endpoint, _task_prefix = "http-wellknown")


class AnvilCookie(object):

    def __init__(self, type):
        self._type = type

    def __getitem__(self, name):
        return _do_call([self._type, name], None, fn_name="anvil.private.get_cookie")

    def __setitem__(self, name, value):
        kw = {}
        kw[name] = value
        self.set(30, **kw)

    def __delitem__(self, name):
        _do_call([self._type, name], None, fn_name="anvil.private.del_cookie")

    def get(self, key, default=None):
        try:
            return _do_call([self._type, key], None, fn_name="anvil.private.get_cookie")
        except KeyError:
            return default

    def set(self, timeout_days=30, **values):
        _do_call([self._type, timeout_days], values, fn_name="anvil.private.set_cookie")

    def clear(self):
        _do_call([self._type], None, fn_name="anvil.private.clear_cookie")


class CookieContainer(object):

    def __init__(self):
        self.local = AnvilCookie("local")
        self.shared = AnvilCookie("shared")


cookies = CookieContainer()


class NotABackgroundTaskState(object):
    def __setitem__(self, key, value):
        raise Exception("Cannot access anvil.server.task_state outside a background task")

    def __getitem__(self, item):
        raise Exception("Cannot access anvil.server.task_state outside a background task")


# Raise event on all sessions in the current environment, on a specific session, on multiple sessions, or on a named channel
def raise_event(name, payload=None, session_id=None, session_ids=None, channel=None):
    anvil.server.call("anvil.private.raise_event", name, payload, session_id=session_id, session_ids=session_ids, channel=channel)


# List all sessions in the current environment, or only those where the specified user is logged in.
def list_client_sessions(user=None):
    return anvil.server.call("anvil.private.list_sessions", user=user)


# Get the value of anvil.server.session for a particular session in the current environment
def get_client_session(session_id):
    return anvil.server.call("anvil.private.get_session_data", session_id)


# Get the ID of the current session
#!defFunction(anvil.server,%)!2: "Returns the current session's ID." ["get_session_id"]
def get_session_id():
    return anvil.server.call("anvil.private.get_session_id")


# Subscribe this session to receive events from the named channel
def subscribe(channel):
    return anvil.server.call("anvil.private.subscribe", channel)


# Unsubscribe this session from receiving events from the named channel
def unsubscribe(channel):
    return anvil.server.call("anvil.private.unsubscribe", channel)


def get_subscriptions():
    return anvil.server.call("anvil.private.get_subscriptions")


def plotly_serialization_helper(class_fullname):
    name_parts = class_fullname.split(".")
    module_name = ".".join(name_parts[:-1])
    class_name = name_parts[-1]

    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)

    if not hasattr(cls, '__serialize__'):
        # print(f"Registering {cls}")
        def serialize(self, global_data):
            # print("Serialising %s on downlink" % type(self))
            return self.to_plotly_json()

        @staticmethod
        def new_deserialized(data, global_data):
            # print("Deserialising %s on downlink" % cls)
            return cls(data)

        cls.__serialize__ = serialize
        cls.__new_deserialized__ = new_deserialized
        portable_class(cls, class_fullname)


_serialization_helpers["plotly.graph_objs"] = plotly_serialization_helper



# Wrap any SQL database in a Data-Tables-like API
try:
    from typing import Dict, Tuple
except ImportError:
    import collections
    Tuple = Dict = collections.defaultdict(dict)
import anvil.server
from anvil.server import Capability


class DatabaseImpl:
    def search(self, table: "TableInfo"):
        raise NotImplementedError

    def add_row(self, table: "TableInfo", data: dict):
        raise NotImplementedError

    def update_row(self, table: "TableInfo", primary_key: tuple, updates: dict):
        raise NotImplementedError

    def delete_row(self, table: "TableInfo", primary_key: tuple):
        raise NotImplementedError


def shared_data(gd, db_name):
    return gd.shared_data("anvil.ext_tables/"+db_name,
                          local_data_factory=lambda: {"TableInfo": {}},
                          remote_data_factory=lambda: {"tables": {}})

@anvil.server.portable_class
class TableInfo:
    def __init__(self, db_name, name, primary_key: Tuple[str], columns, db: DatabaseImpl):
        self.db_name = db_name
        self.name = name
        self.primary_key = primary_key
        self.columns = columns
        self.db = db

    @property
    def safe_version(self):
        return self

    def __serialize__(self, gd):
        return self.db_name, self.name, self.primary_key, self.columns

    def __deserialize__(self, value, gd):
        self.db_name, self.name, self.primary_key, self.columns = value
        self.db = None
    # def __serialize__(self, gd):
    #     txdata, _ = shared_data(gd, self.db_name)
    #     txdata["tables"][self.name] = (self.primary_key, self.columns)
    #     return self.db_name, self.name
    #
    # @classmethod
    # def __new_deserialized__(cls, data, gd):
    #     db_name, name = data
    #     txdata, localdata = shared_data(gd, db_name)
    #     inst = localdata["TableInfo"].get(name)
    #     if not inst:
    #         primary_key, columns = txdata["tables"][name]
    #         inst = cls(db_name, name, primary_key, columns, None)
    #     return inst


@anvil.server.portable_class
class Table:
    def __init__(self, info: TableInfo):
        self.info = info
        self.cap = anvil.server.Capability(["ext_tables/"+info.db_name, info.name])

    def search(self):
        if self.info.db:
            return self.info.db.search(table=self.info)
        else:
            return anvil.server.call("anvil.ext_tables.search/"+self.db_name, cap=self.cap)

    def add_row(self, **data):
        if self.info.db:
            return self.info.db.add_row(table=self.info, data=data)
        else:
            return anvil.server.call("anvil.ext_tables.add_row/"+self.db_name, cap=self.cap, data=data)


@anvil.server.portable_class
class Row:
    def __init__(self, table: TableInfo, data: dict, cap: Capability=None):
        self.table = table
        self.data = data
        self.cap = cap or Capability(["ext_tables/"+table.db_name, table.name, self.primary_key])

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.update(**{key: value})

    def __serialize__(self, gd):
        return {"table": self.table.safe_version, "data": self.data, "cap": self.cap}

    @property
    def primary_key(self):
        return tuple(self.data[col] for col in self.table.primary_key)

    def update(self, **updates):
        if self.table.db:
            self.table.db.update_row(self.table, self.primary_key, updates)
        else:
            anvil.server.call("anvil.ext_tables.update_row/"+self.table.db_name, cap=self.cap, updates=updates)
        self.data.update(updates)

    def delete(self):
        if self.table.db:
            self.table.db.delete_row(self.table, self.primary_key)
        else:
            anvil.server.call("anvil.ext_tables.delete_row/"+self.table.db_name, cap=self.cap)



import time

import anvil.server

from ._base_classes import Row, SearchIterator, Table
from . import _config
from ._errors import NoSuchColumnError, QuotaExceededError, RowDeleted, TableError, TransactionConflict
from ._helpers import _hash_wrapper

# Hack: Force ourselves into the top-level package, even
# if we were loaded into a runtime-v1 per-app Anvil package
__package__ = "anvil.tables"
__name__ = "anvil.tables"


# Use old app tables by default
class AppTables(object):
    cache = None

    def __getattr__(self, name):
        if AppTables.cache is None:
            AppTables.cache = anvil.server.call("anvil.private.tables.get_app_tables")

        tbl = AppTables.cache.get(name)
        if tbl is not None:
            return tbl

        raise AttributeError("No such app table: '%s'" % name)

    def __setattr__(self, name, val):
        raise Exception("app_tables is read-only")
    
    def __iter__(self):
        return AppTableIterator()
    

class AppTableIterator:
    def __init__(self):
        self._it = None

    def __iter__(self):
        return self
    
    def __next__(self):
        # because __iter__ can't suspend
        if AppTables.cache is None:
            AppTables.cache = anvil.server.call("anvil.private.tables.get_app_tables")
        if self._it is None:
            self._it = AppTables.cache.keys().__iter__()
        return next(self._it)
    
    next = __next__


_set_class = object.__dict__["__class__"].__set__

def _lazy_replace_class(self):
    if _config.get_client_config().get("enable_v2"):
        from . import v2
        v2._app_tables._clear_cache()
        _set_class(self, type(v2.app_tables))
    else:
        AppTables.cache = None
        _set_class(self, AppTables)



def _wrap_dunder(method):
    def wrapped(self, *args, **kws):
        _lazy_replace_class(self)
        return getattr(self, method)(*args, **kws)

    return wrapped


class _LazyAppTables(object):
    def __getattribute__(self, name):
        _lazy_replace_class(self)
        return getattr(self, name)

    __setattr__ = _wrap_dunder("__setattr__")
    __getitem__ = _wrap_dunder("__getitem__")
    __dir__ = _wrap_dunder("__dir__")
    __iter__ = AppTables.__iter__


class _LazyContext(object):
    def __enter__(self):
        global batch_update, batch_delete
        if not _config.get_client_config().get("enable_v2"):
            batch_update.__class__ = batch_delete.__class__ = type(None)
            return self.__enter__()

        from .v2 import _batcher as _b

        for obj, orig in zip((batch_update, batch_delete), ("batch_update", "batch_delete")):
            obj.__class__ = type(getattr(_b, orig))
            obj.__init__()
            setattr(_b, orig, obj)
        return self.__enter__()

    def __exit__(self, *args):
        assert not isinstance(self, _LazyContext)
        return self.__exit__(*args)


def _clear_cache():
    _config.reset_config()
    _set_class(app_tables, _LazyAppTables)


#!defModuleAttr(anvil.tables)!1:
# {
# 	name: "app_tables",
# 	type: "any",
# 	anvil$helpLink: "/docs/data-tables/data-tables-in-code",
# 	$doc: "Access Table objects from the datatables services. You can access a Table object with dot notation e.g. `app_tables.my_table`. To access a table with strings use `getattr(app_tables, 'my_table')`. If no table is present an AttributeError will be thrown."
# }
#
app_tables = _LazyAppTables()
batch_update = _LazyContext()
batch_delete = _LazyContext()
# Not very nice but these references exist in uplink code
# before we have a chance to know if we're using the v1/v2 config option
# we can't call anvil.server until the uplink has made a connetion


def get_table_by_id(table_id):
    if _config.get_client_config().get("enable_v2"):
        from .v2 import get_table_by_id

        return get_table_by_id(table_id)
    raise TableError("get_table_by_id is only available in Accelerated Tables beta")


#!defModuleAttr(anvil.tables)!1:
# {
# 	name: "app_tables",
# 	type: "any",
# 	anvil$helpLink: "/docs/data-tables/data-tables-in-code",
# 	$doc: "Access Table objects from the datatables services. You can access a Table object with dot notation e.g. `app_tables.my_table`. To access a table with strings use `getattr(app_tables, 'my_table')`. If no table is present an AttributeError will be thrown."
# }
#


class Transaction:
    def __init__(self, relaxed=False):
        self._aborting = False
        self._isolation = "relaxed" if relaxed else None

    #!defMethod(anvil.tables.Transaction instance)!2: "Begin the transaction" ["__enter__"]
    def __enter__(self):
        anvil.server.call("anvil.private.tables.open_transaction", isolation=self._isolation)
        return self

    #!defMethod(_)!2: "End the transaction" ["__exit__"]
    def __exit__(self, e_type, e_val, tb):
        anvil.server.call("anvil.private.tables.close_transaction", self._aborting or e_val is not None)

    #!defMethod(_)!2: "Abort this transaction. When it ends, all write operations performed during it will be cancelled" ["abort"]
    def abort(self):
        self._aborting = True


#!defClass(anvil.tables,%Transaction)!:


#!defFunction(anvil.tables,%,function,server_function)!2:
# {
# 	$doc: "When applied to a function (as a decorator), the whole function will run in a data tables transaction. If it conflicts with another transaction, it will retry up to five times.",
# anvil$helpLink: "/docs/data-tables/transactions"
#  } ["in_transaction"]
def in_transaction(maybe_f=None, relaxed=None):
    def wrap(f):
        def new_f(*args, **kwargs):
            n = 0
            while True:
                try:
                    with Transaction(relaxed=relaxed):
                        return f(*args, **kwargs)
                except TransactionConflict:
                    # lazy load random incase we make random.js a slow path on the client
                    import random

                    n += 1
                    if n == 18:
                        raise
                    # print(f"RETRYING TXN {n}")
                    # Max total sleep time is a little under 150 seconds (avg 75), so server calls will timeout before this finishes usually.
                    sleep_amt = random.random() * (1.5**n) * 0.05
                    try:
                        time.sleep(sleep_amt)
                    except:
                        anvil.server.call("anvil.private._sleep", sleep_amt)

        try:
            reregister = f._anvil_reregister
        except AttributeError:
            pass
        else:
            reregister(new_f)

        new_f.__name__ = f.__name__

        return new_f

    if maybe_f is None:
        return wrap
    else:
        return wrap(maybe_f)


#!defFunction(anvil.tables,_,column_name,ascending=)!2: "Sort the results of this table search by a particular column. Default to ascending order." ["order_by"]
@anvil.server.portable_class
class order_by(object):
    def __init__(self, column_name, ascending=True):
        self.column_name = column_name
        self.ascending = ascending

    __hash__, __eq__ = _hash_wrapper("column_name", "ascending")


# backward compatability
from .query import fetch_only
from .query import page_size as _page_size


#!defFunction(anvil.tables,%,[via_host=],[via_port=])!2: "Get a Postgres connection string for accessing this app's Data Tables via SQL.\n\nThe returned string includes temporary login credentials and sets the search path to a schema representing this app's Data Table environment.\n\nYou can override the host and port for the database connection to connect via a secure tunnel.\n\n(Available on the Dedicated Plan only.)" ["get_connection_string"]
def get_connection_string(via_host=None, via_port=None):
    return anvil.server.call(
        "anvil.private.get_direct_postgres_connection_string", via_host=via_host, via_port=via_port
    )


#!defMethod(table row, **column_values)!2: "Add a row to the data table. Use keyword arguments to specify column values." ["add_row"]
#!defMethod(client readable view)!2: "Return a view on the table that can be read by client code. Use keyword arguments to specify view restrictions" ["client_readable"]
#!defMethod(client writable view)!2: "Return a view on the table that can be written by client code. Use keyword arguments to specify view restrictions. This does not give the client write access to other tables referred to by the table." ["client_writable"]
#!defMethod(client writable view)!2: "Return a view on this table that can be written by client code. Use keyword arguments to specify view restrictions." ["client_writable_cascade"]
#!defMethod(_)!2: "Delete all the rows from the data table" ["delete_all_rows"]
#!defMethod(_)!2: "Get a single matching row from the data table whose columns match the keyword arguments. Returns None if no matching row exists, and raises an exception if more than one row matches.\n\nEg: app_tables.table_1.get(name='John Smith')" ["get"]
#!defMethod(row,id)!2: "Get the matching row from this data table, by its unique ID" ["get_by_id"]
#!defMethod(bool,row)!2: "Returns true if the table (or view) contains the provided row." ["has_row"]
#!defMethod(list of dicts)!2: "Get the spec for the table as a list of dicts. Each dict contains the name and type of a column." ["list_columns"]
#!defMethod(Row or None)!2: "Get rows from a data table. If you specify keyword arguments, you will retrieve only rows whose columns match those values.\n\nEg: app_tables.table_1.search(name='John Smith')" ["search"]
#!defMethod(Media object, [escape_for_excel=False])!2: "Get the table in CSV format, optionally escaped for use in Excel. Returns a downloadable Media object; use its url property." ["to_csv"]
#!defClassNoConstructor(anvil.tables,#Table)!1: "A table returned from app_tables"

#!defMethod(Media object, [escape_for_excel=False])!2: "Get the results of the SearchIterator in CSV format, optionally escaped for use in Excel. Returns a downloadable Media object; use its url property." ["to_csv"]
#!defClassNoConstructor(anvil.tables,#SearchIterator)!1: "An iterator of table rows returned from a search()";


#!defMethod(_)!2: "Delete the row from its data table" ["delete"]
#!defMethod(id)!2: "Get the unique ID of the table row" ["get_id"]
#!defMethod(_,**column_values)!2: "update the data for multiple columns" ["update"]
#!defClassNoConstructor(anvil.tables,#Row)!1: "A table row";



import anvil.server

from .._base_classes import AppTables as BaseAppTables
from ._constants import SERVER_PREFIX
from ._table import Table

_table_cache = None


def _fill_cache():
    global _table_cache
    if _table_cache is None:
        _table_cache = anvil.server.call(SERVER_PREFIX + "get_app_tables")
    return _table_cache


def _clear_cache():
    global _table_cache
    _table_cache = None


class AppTableIterator:
    def __init__(self):
        self._it = None

    def __iter__(self):
        return self
    
    def __next__(self):
        if self._it is None:
            self._it = _fill_cache().__iter__()
        return next(self._it)
    
    next = __next__


class AppTables(BaseAppTables):
    def __getattribute__(self, name):
        # use __getattribute__ so that we prioritise the table name
        try:
            return self[name]
        except KeyError:
            return object.__getattribute__(self, name)

    def __getitem__(self, name):
        cache = _fill_cache()
        table_args = cache[name]
        return Table._create(*table_args)

    def __setattr__(self, name, val):
        raise AttributeError("app_tables is read-only")

    def __dir__(self):
        return object.__dir__(self) + list(_fill_cache().keys())
    
    def __iter__(self):
        return AppTableIterator()



def get_table_by_id(table_id):
    table_args = anvil.server.call(SERVER_PREFIX + "get_table_by_id", table_id)
    return table_args and Table._create(*table_args)


app_tables = AppTables()
